import os

file_content = ["# flake8: noqa"]

for entry in os.scandir("docs/examples"):
    if entry.is_file():
        file_name_split = entry.name.split(".")
        if file_name_split[-1] == "py":
            file_content.append(
                f"def test_{file_name_split[0]}():\n\tfrom docs.examples import {file_name_split[0]} # noqa"
            )

    with open("tests/test_examples.py", "w") as f:
        f.write("\n\n".join(file_content))



from typing import TYPE_CHECKING, Iterable, List, Union

import pandas as pd
from rich import box
from rich.console import Group
from rich.layout import Layout
from rich.panel import Panel
from rich.table import Table

from krisi.utils.iterable_helpers import group_by_categories
from krisi.utils.printing import (
    create_metric_table,
    create_y_pred_table,
    metrics_empty_in_category,
)

if TYPE_CHECKING:
    from krisi.evaluate.metric import Metric
    from krisi.evaluate.scorecard import ScoreCard

from krisi.utils.printing import bold


def get_metric_string(obj: "Metric", repr: bool = False) -> str:
    if repr:
        return f"{obj.result} | {obj.name}"
    hyperparams = ""
    if obj.parameters is not None:
        hyperparams += "".join(
            [f"{key} - {value}" for key, value in obj.parameters.items()]
        )
    if (
        obj.result is None
        and obj.result_rolling is not None
        and isinstance(obj.result_rolling, Iterable)
    ):
        result_ = (
            "[" + ", ".join([f"{result:<0.5}" for result in obj.result_rolling]) + "]"
        )
    elif obj.result is None:
        result_ = ""
    elif isinstance(obj.result, Iterable) and not isinstance(obj.result, str):
        result_ = obj.result
    else:
        result_ = f"{obj.result:<15.5}"

    return f"{obj.name:>40s} ({obj.key}): {result_}{hyperparams:>15s}"


def get_minimal_summary(obj: "ScoreCard", dataframe: bool) -> Union[pd.DataFrame, str]:
    if dataframe:
        all_metrics = [
            metric
            for metric in obj.get_all_metrics()
            if isinstance(metric.result, (float, int))
        ]
        return pd.Series(
            [metric.result for metric in all_metrics],
            name=obj.metadata.model_name,
            index=[f"{metric.name:>40s}" for metric in all_metrics],
        ).to_frame()

    return "\n".join(
        [
            f"{metric.name:>40s} - {metric.result:<15.5}"
            for metric in obj.get_all_metrics()
            if isinstance(metric.result, (float, int))
        ]
    )


def get_large_metric_summary(
    obj: "ScoreCard",
    title: str,
    with_info: bool,
    with_parameters: bool,
    with_diagnostics: bool,
) -> Table:
    return create_metric_table(
        title=title,
        metrics=obj.get_all_metrics(),
        with_info=with_info,
        with_parameters=with_parameters,
        with_diagnostics=with_diagnostics,
    )


def get_summary(
    obj: "ScoreCard",
    categories: List[str],
    repr: bool = True,
    with_info: bool = False,
    with_parameters: bool = True,
    with_diagnostics: bool = False,
    input_analysis: bool = True,
) -> Union[Panel, Layout]:
    category_groups = group_by_categories(list(vars(obj).values()), categories)
    input_analysis_table = (
        [
            create_y_pred_table(
                obj.dataset_type.is_classification(), obj.y, obj.predictions
            )
        ]
        if input_analysis
        else []
    )

    metric_tables = Group(
        *input_analysis_table
        + [
            create_metric_table(
                f"{category if category is not None else 'Unknown Category':>15s}",
                metrics,
                with_info=with_info,
                with_parameters=with_parameters,
                with_diagnostics=with_diagnostics,
                show_header=True if index == 0 else False,
            )
            for index, (category, metrics) in enumerate(category_groups.items())
            if not metrics_empty_in_category(metrics)
        ],
    )

    title = f"Result of {obj.metadata.model_name if repr else bold(obj.metadata.model_name)} on {obj.metadata.dataset_name if repr else bold(obj.metadata.dataset_name)} tested on {obj.sample_type.value if repr else bold(obj.sample_type.value)}"
    return Panel(metric_tables, title=title, padding=1, box=box.HEAVY_EDGE, expand=True)



from typing import Any, Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from typing_extensions import get_args, get_origin

from krisi.evaluate.metric import Metric
from krisi.evaluate.type import DatasetType, Predictions, Targets, Weights
from krisi.utils.iterable_helpers import flatten, isiterable

valid_types = ["int64", "float64", "float32", "float", "int"]


def get_bad_types(iterable: Union[Targets, Predictions]) -> Iterable:
    return [x for x in iterable if type(x) not in valid_types]


def flatten_type(complex_type: Any) -> Union[List, Any]:
    types = get_args(complex_type)
    if len(types) == 0:
        return complex_type
    else:
        add_list = []
        if isinstance(get_origin(complex_type), type([])):
            add_list = [type([])]
        return add_list + [flatten_type(type_) for type_ in get_args(complex_type)]


def unpack_type(complex_type: Any) -> Tuple[Any]:
    return tuple(flatten(flatten_type(complex_type)))


def check_no_duplicate_metrics(metrics: List[Metric]):
    assert len(metrics) == len(
        set([metric.key for metric in metrics])
    ), f"Duplicate metric key found in {metrics}. Please remove the duplicates."


def check_valid_pred_target(
    y: Targets, predictions: Predictions, sample_weights: Optional[Weights]
):
    assert len(y) == len(
        predictions
    ), f"Target length {len(y)} should match predictions length {len(predictions)}"
    if sample_weights is not None:
        assert len(y) == len(
            sample_weights
        ), f"Target length {len(y)} should match sample_weights length {len(sample_weights)}"

    assert isinstance(
        y, unpack_type(Targets)
    ), f"Targets: {type(y)} are not of a valid type."
    assert isinstance(
        predictions, unpack_type(Predictions)
    ), f"Predictions: {type(predictions)} are not of a valid type."

    for iterable in [y, predictions]:
        assert isiterable(iterable), f"{iterable} is not an iterable or it is a string."
        dtype_name = ""
        if isinstance(iterable, pd.Series):
            dtype_name = iterable.dtype
        elif isinstance(iterable, np.ndarray):
            dtype_name = iterable.dtype.name
        else:
            types = set([type(obj) for obj in iterable])
            assert len(types) == 1, f"Inconsistent types in {iterable}"
            dtype_name = list(types)[0]

        assert (
            dtype_name in valid_types
        ), f"{iterable} contains at least one invalid type. {get_bad_types(iterable)}"


def infer_dataset_type(y: Targets) -> DatasetType:
    # TODO: Should work with booleans
    def infer_exact_type(is_classification: bool, y: Targets) -> DatasetType:
        if is_classification is False:
            return DatasetType.regression
        if len(set(y)) == 2:
            class_1 = y[0]
            no_of_class_1 = np.count_nonzero(y == class_1)
            ratio = no_of_class_1 / len(y)
            if ratio < 0.3 or ratio > 0.7:
                return DatasetType.classification_binary_imbalanced
            else:
                return DatasetType.classification_binary_balanced
        else:
            return DatasetType.classification_multiclass

    if isinstance(y, pd.Series):
        y = y.to_numpy()
    if isinstance(y, np.ndarray):
        if hasattr(y[0], "is_integer"):
            # Using this is faster, if available use this.
            return infer_exact_type(
                all([i.is_integer() for i in y if i is not None]), y
            )
        else:
            return infer_exact_type(np.all(np.mod(y, 1) == 0), y)
    else:
        return infer_exact_type(
            all([isinstance(i, int) for i in y if i is not None]), y
        )



import os
import sys


def is_testing() -> bool:
    in_testing = False
    if "PYTEST_CURRENT_TEST" in os.environ:
        in_testing = True

    for argv_ in sys.argv:
        script_name = os.path.basename(argv_)
        if script_name in ["pytest", "py.test"]:
            in_testing = True

    return in_testing



from time import monotonic
from typing import Any, Callable, Tuple

import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error

from krisi import score
from krisi.evaluate.type import Calculation

np.random.seed(42)
func = mean_absolute_error
lengthofdataset = 1000000

y = pd.Series(np.random.rand(lengthofdataset), name="y")
predictions = pd.Series(np.random.rand(lengthofdataset), name="predictions")


window = 30


def timeingit(func: Callable) -> Tuple[Any, float]:
    start_time = monotonic()
    res = func()
    runtime = monotonic() - start_time
    print(f"Run time: {runtime} seconds")
    return res, runtime


def func1():
    return [
        func(y_roll, predictions[y_roll.index])
        for y_roll in y.rolling(window=window, step=window)
    ]


def func2():
    return [
        func(single_window["y"], single_window["predictions"])
        for single_window in pd.concat([y, predictions], axis="columns").rolling(
            window=window, step=window
        )
    ]


def func3():
    return [
        func(y[i : i + window], predictions[i : i + window])
        for i in range(0, len(y) - 1, window)
    ]


def func4():
    return [
        func(*single_window.values.T)
        for single_window in pd.concat([y, predictions], axis="columns").rolling(
            window=window, step=window
        )
    ]


def func_actual():
    score(
        y=np.random.rand(lengthofdataset),
        predictions=np.random.rand(lengthofdataset),
        calculation=Calculation.rolling,
        rolling_args={"window": lengthofdataset // 100, "step": lengthofdataset // 100},
    )


# res1 = timeingit(func1)
# res2 = timeingit(func2)
# # res3 = timeingit(func3)
# res4 = timeingit(func4)
# timeingit(func_actual)

# assert res1 == res2 == res4
# assert res1 == res2 == res3 == res4



from rich.ansi import AnsiDecoder
from rich.console import Group
from rich.jupyter import JupyterMixin


class plotextMixin(JupyterMixin):
    def __init__(self, data, plotting_function, title=""):
        self.decoder = AnsiDecoder()
        self.title = title
        self.data = data
        self.plotting_function = plotting_function

    def __rich_console__(self, console, options):
        self.width = options.max_width or console.width
        self.height = options.height or console.height
        canvas = self.plotting_function(self.data, self.width, self.height, self.title)
        self.rich_canvas = Group(*self.decoder.decode(canvas))
        yield self.rich_canvas



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

import pandas as pd
from typing_extensions import Literal

from krisi.utils.enums import ParsableEnum

if TYPE_CHECKING:
    import plotly.graph_objects as go

from krisi.evaluate.type import MetricCategories, MetricResult

PlotFunction = Callable[
    [List[MetricResult], str, Optional[float]],
    "go.Figure",
]
PlotDefinition = Tuple[PlotFunction, Optional[Dict[str, Any]]]


@dataclass
class PlotlyInput:
    type: Literal[
        "dcc.Dropdown", "dcc.Input", "dcc.Checklist", "dcc.Slider", "dcc.RangeSlider"
    ]
    id: str
    value_name: str
    default_value: Union[str, float, int]
    options: Optional[List[Union[str, float, int]]] = None


@dataclass
class InteractiveFigure:
    id: str
    get_figure: Callable
    inputs: List[PlotlyInput] = field(default_factory=list)
    global_input_ids: List[str] = field(default_factory=list)
    title: Optional[str] = ""
    width: Optional[float] = 900.0
    height: Optional[float] = 600.0
    category: Optional[MetricCategories] = None
    plot_args: Dict[str, Any] = field(default_factory=dict)


class DisplayModes(ParsableEnum):
    interactive = "interactive"
    pdf = "pdf"
    direct = "direct"
    direct_save = "direct_save"
    direct_one_subplot = "direct_one_subplot"


def plotly_interactive(
    plot_function: PlotFunction,
    data_source: Union[pd.DataFrame, pd.Series],
    *args,
    **kwargs,
) -> Callable:
    # default_args = args
    default_kwargs = kwargs

    def wrapper(*args, **kwargs) -> "go.Figure":
        for key, value in default_kwargs.items():
            if key not in kwargs:
                kwargs[key] = value

        width = kwargs.pop("width", None)
        height = kwargs.pop("height", None)
        title = kwargs.get("title", None)

        fig = plot_function(data_source, *args, **kwargs)
        fig.update_layout(autosize=False, width=width, height=height)
        if title is not None:
            fig.update_layout(title=title)
        return fig

    return wrapper



import datetime
import os
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Tuple, Union

from krisi.evaluate.type import MetricCategories, PathConst, ScoreCardMetadata
from krisi.report.interactive import run_app
from krisi.report.pdf import (
    convert_figure_to_html,
    convert_figures_to_html,
    convert_figures_to_side_by_side_html,
    create_pdf_report,
)
from krisi.report.type import DisplayModes, InteractiveFigure, PlotlyInput
from krisi.report.vizualise import create_subplots_from_mutiple_plots
from krisi.utils.io import ensure_path
from krisi.utils.iterable_helpers import (
    flatten,
    group_by_categories,
    remove_nans,
    replace_if_none,
    wrap_in_list,
)
from krisi.utils.state import RunType, get_global_state

if TYPE_CHECKING:
    from krisi.evaluate.metric import Metric
    from krisi.evaluate.scorecard import ScoreCard


class Report:
    title: str
    modes: List[Union[str, DisplayModes]]

    def __init__(
        self,
        title: str,
        general_description: str = "",
        modes: Union[str, List[str], DisplayModes, List[DisplayModes]] = [
            DisplayModes.pdf
        ],
        figures: List[InteractiveFigure] = [],
        global_controllers: List[PlotlyInput] = [],
        html_template_url: Path = PathConst.html_report_template_url,
        css_template_url: Path = Path("library/default/template.css"),
        get_html_elements: Optional[Callable] = None,
        scorecard_metadata: Optional[ScoreCardMetadata] = None,
        save_path: Optional[Path] = None,
    ) -> None:
        self.title = title
        self.general_description = general_description
        self.scorecard_metadata = scorecard_metadata

        self.modes = [DisplayModes.from_str(mode) for mode in wrap_in_list(modes)]

        self.figures = figures
        self.global_controllers = global_controllers
        self.html_template_url = html_template_url
        self.css_template_url = css_template_url
        self.get_html_elements = get_html_elements
        self.save_path = replace_if_none(
            save_path,
            Path(os.path.join(PathConst.default_eval_output_path, Path(title))),
        )

    def generate_launch(self) -> None:
        figures_by_category = group_by_categories(
            self.figures, [el.value for el in MetricCategories]
        )

        if DisplayModes.pdf in self.modes:
            create_pdf_report(
                path=self.save_path,
                title=self.title,
                html_template_url=self.html_template_url,
                css_template_url=self.css_template_url,
                html_elements_to_inject=self.get_html_elements()
                if self.get_html_elements is not None
                else dict(),
            )

        if DisplayModes.direct in self.modes:
            if not get_global_state().run_type == RunType.test:
                [
                    figure.get_figure(**figure.plot_args).show()
                    for figure in self.figures
                ]

        if DisplayModes.direct_save in self.modes:
            diagram_path = Path(os.path.join(self.save_path, Path("diagrams")))
            ensure_path(diagram_path)
            for i, figure in enumerate(self.figures):
                figure.get_figure(**figure.plot_args).write_image(
                    os.path.join(diagram_path, f"{i}_{figure.title}.svg")
                )

        if DisplayModes.direct_one_subplot in self.modes:
            if not get_global_state().run_type == RunType.test:
                create_subplots_from_mutiple_plots(
                    self.figures, title=self.title
                ).show()

        if DisplayModes.interactive in self.modes:
            run_app(
                figures_by_category,
                self.global_controllers,
                self.title,
                self.general_description,
                self.scorecard_metadata,
            )


def get_all_interactive_diagrams(metrics: List["Metric"]) -> List[InteractiveFigure]:
    interactive_diagrams = remove_nans(
        [metric.get_diagrams() for metric in metrics]
        + [metric.get_diagram_over_time() for metric in metrics]
    )
    return flatten(interactive_diagrams)


def get_group_html(
    group_name: str, metrics: List["Metric"]
) -> Tuple[str, List[InteractiveFigure]]:
    html_images = f'<section class="mb-6"><br><h3 class="mb-6 text-sm font-bold">{group_name}</h3><br>'
    interactive_diagrams = get_all_interactive_diagrams(metrics)

    interactive_diagrams.sort(key=lambda x: x.plot_args["width"], reverse=True)

    for i in range(len(interactive_diagrams)):
        if (
            i + 1 != len(interactive_diagrams)
            and interactive_diagrams[i].plot_args["width"] < 1000.0
            and interactive_diagrams[i + 1].plot_args["width"] < 1000.0
        ):
            html_images += convert_figures_to_side_by_side_html(
                [
                    interactive_diagrams[i].get_figure(
                        **interactive_diagrams[i].plot_args
                    ),
                    interactive_diagrams[i + 1].get_figure(
                        **interactive_diagrams[i + 1].plot_args
                    ),
                ]
            )
        else:
            html_images += convert_figure_to_html(
                interactive_diagrams[i].get_figure(**interactive_diagrams[i].plot_args)
            )

    html_images += "</section>"

    return html_images, interactive_diagrams


def get_waterfall_metric_html(
    metrics: List["Metric"],
) -> Tuple[str, List[InteractiveFigure]]:
    category_groups = group_by_categories(
        metrics, categories=[el.value for el in MetricCategories]
    )
    html_and_diagrams = [
        get_group_html(group_name, metrics)
        for group_name, metrics in category_groups.items()
        if metrics is not None and len(metrics) > 0 and group_name is not None
    ]
    return "".join([html for html, _ in html_and_diagrams]), flatten(
        [diagrams for _, diagrams in html_and_diagrams]
    )


def get_html_elements_for_injection_scorecard(
    obj: "ScoreCard",
    author: str,
    project_name: str,
    date: str,
    custom_metric_html: str,
) -> Callable:
    def func() -> Dict[str, str]:
        diagrams = obj.get_diagram_dictionary()

        diagrams_static = {
            interactive_figure.id: convert_figures_to_html(
                [
                    interactive_figure.get_figure(
                        title=interactive_figure.title,
                        **interactive_figure.plot_args,
                    )
                ]
            )
            for key, interactive_figure in diagrams.items()
        }

        return dict(
            author=author,
            project_name=project_name,
            date=date,
            custom_metric_html=custom_metric_html,
            **diagrams_static,
        )

    return func


def create_report_from_scorecard(
    obj: "ScoreCard",
    display_modes: List[DisplayModes],
    html_template_url: Path,
    css_template_url: Path,
    author: str,
    report_title: str,
    save_path: Path,
) -> Report:
    if DisplayModes.pdf in display_modes or DisplayModes.pdf.value in display_modes:
        custom_metric_html, interactive_diagrams = get_waterfall_metric_html(
            obj.get_all_metrics()
        )
        get_html_elements = get_html_elements_for_injection_scorecard(
            obj=obj,
            author=author,
            project_name=obj.metadata.project_name,
            date=datetime.datetime.now().strftime("%Y-%m-%d"),
            custom_metric_html=custom_metric_html,
        )
    else:
        get_html_elements = None
        interactive_diagrams = get_all_interactive_diagrams(obj.get_all_metrics())
        interactive_diagrams.sort(key=lambda x: x.plot_args["width"], reverse=True)

    return Report(
        title=report_title,
        general_description="General Description",
        modes=display_modes,
        scorecard_metadata=obj.metadata,
        figures=interactive_diagrams,
        html_template_url=html_template_url,
        css_template_url=css_template_url,
        get_html_elements=get_html_elements,
        save_path=save_path,
    )



from dataclasses import dataclass

from krisi.utils.enums import ParsableEnum


class RunType(ParsableEnum):
    dev = "dev"
    test = "test"
    prod = "prod"


@dataclass
class GlobalState:
    run_type: RunType = RunType.dev
    verbose: bool = True


GLOBAL_STATE = GlobalState(run_type=RunType.dev)


def set_global_state(state: GlobalState):
    global GLOBAL_STATE
    GLOBAL_STATE = state


def get_global_state() -> GlobalState:
    return GLOBAL_STATE



import itertools
from random import sample
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from krisi.sharedtypes import Task


def make_it_stationary(ds: pd.Series) -> pd.Series:
    """If the ds, is a random walk with drift, take first differences to make it stationary."""
    print(f"Making the Series {ds.name} stationary.")
    return ds.diff().dropna()


def generating_arima_synthetic_data(
    target_col: str,
    nsample: int = 1000,
    ar: List[float] = [1.0, -0.9],
    ma: List[float] = [1.0],
) -> pd.Series:
    from statsmodels.tsa.arima_process import ArmaProcess

    """Generate Synthetic data with ARIMA process."""
    print(f"Generating Synthetic ARIMA process with nsamples:{nsample}")

    # Plot 1: AR parameter = +0.9,
    # exponent of PHI is opposite due to original signal processing literature convention
    ar = np.array(ar)
    ma = np.array(ma)
    AR_object = ArmaProcess(ar, ma)
    simulated_data = AR_object.generate_sample(nsample=nsample)

    return pd.Series(
        simulated_data,
        name=target_col,
        index=pd.date_range(end="2022", periods=nsample, freq="D"),
    )


def create_probabilities(num_labels: int = 3, num_samples: int = 1000):
    probabilities = np.random.rand(num_samples, num_labels)
    probabilities = probabilities / np.sum(probabilities, axis=1, keepdims=True)
    probabilities = list(zip(*probabilities))
    probabilities = pd.DataFrame({i: probabilities[i] for i in range(num_labels)})
    return probabilities


def generate_random_classification(
    num_labels: int = 3, num_samples: int = 1000
) -> Tuple[pd.Series, pd.Series, pd.DataFrame, pd.Series]:
    y = pd.Series(np.random.randint(0, num_labels, num_samples))
    predictions = pd.Series(np.random.randint(0, num_labels, num_samples))
    sample_weight = pd.Series(np.random.random(num_samples))

    probabilities = create_probabilities(num_labels=num_labels, num_samples=num_samples)
    return y, predictions, probabilities, sample_weight


def generate_synthetic_data(
    num_features_increasingly_important: int = 1,  # component, increasing importance, added to `y`
    num_features_decreasingly_important: int = 1,  # component, decreasing importance, added to `y`
    num_features_stationary: int = 1,  # component + autoregressive transformation, static importance, added to `y`
    num_features_correlated_noise: int = 1,  # component, static importance, added to `y`
    num_features_uncorrelated_noise: int = 1,  # pure noise, added as features ONLY to `X`
    num_features_oscillates_importance: int = 1,  # component, oscilating importance, added to `y`
    num_obs: int = 200,
    index: Optional[pd.Index] = None,
    task: Union[str, Task] = Task.regression,
):
    task = Task.from_str(task)
    # Based on https://github.com/convergenceIM/alpha-scientist/blob/master/content/04_Walk_Forward_Modeling.ipynb
    if index is None:
        index = pd.RangeIndex(0, num_obs)

    def add_memory(s, n_days=50, memory_strength=0.1):
        """adds autoregressive behavior to series of data"""
        out = (1 - memory_strength) * s + memory_strength * s.ewm(n_days).mean()
        return out

    class Relevance:
        increasingly_important = np.linspace(0.5, 1.5, num_obs)
        decreasingly_important = np.linspace(1.5, 0.5, num_obs)
        oscillates_importance = pd.Series(
            np.sin(2 * np.pi * np.linspace(0, 1, num_obs) * 2) + 1, index=index
        )
        stationary = 1.0

    features = [
        (
            "increasingly_important",
            num_features_increasingly_important,
            Relevance.increasingly_important,
        ),
        (
            "decreasingly_important",
            num_features_decreasingly_important,
            Relevance.decreasingly_important,
        ),
        (
            "oscillates_importance",
            num_features_oscillates_importance,
            Relevance.oscillates_importance,
        ),
        ("stationary", num_features_stationary, Relevance.stationary),
    ]

    # generate feature data
    def get_feature(name: str, relevance) -> pd.Series:
        feature = pd.Series(np.random.randn(num_obs), index=index, name=name)
        feature = add_memory(feature, 10, 0.1)
        feature *= relevance
        return feature

    def get_features(
        feature_type: str, num_feature: int, relevance: Relevance
    ) -> pd.DataFrame:
        df = pd.concat(
            [get_feature(f"{feature_type}_{i}", relevance) for i in range(num_feature)],
            axis="columns",
        )
        return df

    features_correlated_noise = pd.concat(
        [
            pd.Series(
                np.random.randn(num_obs) * 3, index=index, name=f"correlated_noise_{i}"
            )
            for i in range(num_features_correlated_noise)
        ],
        axis="columns",
    )
    X = pd.concat(
        [
            get_features(feature_type, num_feature, relevance)
            for feature_type, num_feature, relevance in features
        ]
        + [features_correlated_noise],
        axis="columns",
    )

    y = (
        X.sum(axis="columns") + np.random.randn(num_obs) * 3
    )  # We add an uncorrelated noise to the target
    y.name = "y"

    X = pd.concat(
        [X]
        + [
            pd.Series(
                np.random.randn(num_obs) * 3,
                index=index,
                name=f"uncorrelated_noise_{i}",
            )
            for i in range(num_features_uncorrelated_noise)
        ],
        axis="columns",
    )

    if task == Task.classification:
        y = (y > y.median()).astype("int")
    elif task == Task.multi_classification:
        label = y.copy()
        label[y <= y.quantile(0.33)] = 0
        label[y >= y.quantile(0.66)] = 2
        label[(y > y.quantile(0.33)) & (y < y.quantile(0.66))] = 1

        y = label

    return X, y


def generate_synthetic_predictions_binary(
    target: pd.Series,
    sample_weights: Optional[pd.Series] = None,
    index: Optional[pd.Index] = None,
    smoothing_window: Optional[int] = None,
) -> pd.DataFrame:
    if index is None:
        index = target.index
    if sample_weights is None:
        sample_weights = pd.Series(np.ones(len(target)), index=index)
    target = target.copy()

    weighted_target = target * sample_weights
    expanding_target_mean = weighted_target.expanding(min_periods=1).mean()
    prob_mean_class_1 = expanding_target_mean.mean()
    prob_std_class_1 = expanding_target_mean.std()
    prob_class_1 = pd.Series(
        np.random.normal(prob_mean_class_1, prob_std_class_1, len(index)).clip(0, 1)
    )

    if smoothing_window is not None:
        prob_class_1 = prob_class_1.rolling(
            window=smoothing_window, min_periods=1
        ).mean()

    prob_class_0 = 1 - prob_class_1
    return pd.DataFrame(
        {
            "predictions_RandomClassifier": (prob_class_1 > prob_class_1.mean())
            .astype("int")
            .values,
            "probabilities_RandomClassifier_0": prob_class_0.values,
            "probabilities_RandomClassifier_1": prob_class_1.values,
        },
        index=index,
    )


def shuffle_df_in_chunks(
    df: pd.DataFrame, chunk_size: Union[int, float]
) -> pd.DataFrame:
    """Shuffles a dataframe by rows in chunks.

    Parameters
    ----------
    df : pd.DataFrame
        Data, whose rows should be shuffled in chunks
    chunk_size : Union[int, float]
        Either a fixed chunk size or a fraction of the total number of rows

    Returns
    -------
    pd.DataFrame
        Shuffled DataFrame by rows in chunks.
    """
    original_index = df.index
    df = df.reset_index(drop=True)
    chunk_size = (
        int(chunk_size * len(df)) if isinstance(chunk_size, float) else chunk_size
    )
    num_rows = df.shape[0]
    num_chunks = num_rows // chunk_size
    shuffled_idx = []
    for i in sample(list(range(num_chunks)), num_chunks):
        start_idx = i * chunk_size
        end_idx = (i + 1) * chunk_size
        shuffled_idx.append(df.index[start_idx:end_idx])

    # If there are any remaining rows, shuffle and append them as well
    if num_rows % chunk_size != 0:
        shuffled_idx.append(df.index[num_chunks * chunk_size :])

    df = df.copy()
    df.index = df.index[np.concatenate(shuffled_idx)]
    df.sort_index(inplace=True)
    df.index = original_index

    return df


def combinatorial_shuffling_in_chunks(
    df: pd.DataFrame, chunk_size: int, select_index: int
) -> pd.DataFrame:
    original_index = df.index
    df = df.reset_index(drop=True)
    chunks = [(i, i + chunk_size) for i in range(0, len(df), chunk_size)]
    if chunks[-1][1] > len(df):
        chunks[-2] = (chunks[-2][0], len(df))
        chunks = chunks[:-1]

    premuted_chunks = list(itertools.permutations(chunks))[
        1:
    ]  # skip the first one, which is the original order

    flatten_ranges = [
        id for start, stop in premuted_chunks[select_index] for id in range(start, stop)
    ]
    df = df.copy()
    df.index = df.index[flatten_ranges]
    df.sort_index(inplace=True)
    df.index = original_index
    return df



from pathlib import Path
from typing import List

import networkx as nx
import numpy as np
import pandas as pd
from typing_extensions import Literal

from krisi.evaluate.type import PathConst
from krisi.utils.io import ensure_path
from krisi.utils.state import RunType, get_global_state


def create_save_graphs(
    df_rolled_corr: List[pd.DataFrame],
    save_or_display: List[Literal["save", "display"]] = ["save", "display"],
    corr_direction: Literal["positive", "negative"] = "positive",
    min_correlation: float = 0.01,
    save_location: Path = PathConst.default_analyse_output_path,
) -> None:
    ensure_path(save_location)
    for i, corr_df in enumerate(df_rolled_corr):
        __display_corr_graph(
            corr_df,
            save_or_display,
            corr_direction,
            min_correlation,
            save_location,
            file_name=str(i),
        )


def __display_corr_graph(
    corr: pd.DataFrame,
    save_or_display: List[Literal["save", "display"]],
    corr_direction: Literal["positive", "negative"],
    min_correlation: float,
    save_location: Path,
    file_name: str = "0",
) -> None:
    feature_node_names = corr.index.values
    matrix_array = np.array(corr)
    matrix_corr = np.matrix(matrix_array, copy=False, dtype=None)

    G = nx.Graph(matrix_corr)
    G = nx.relabel_nodes(G, lambda i: feature_node_names[i])
    G.remove_edges_from(nx.selfloop_edges(G))

    __create_corr_network(
        G,
        corr_direction=corr_direction,
        min_correlation=min_correlation,
        save_location=save_location,
        file_name=file_name,
        save_or_display=save_or_display,
    )


def __create_corr_network(
    G: nx.Graph,
    corr_direction: Literal["positive", "negative"],
    min_correlation: float,
    save_location: Path,
    file_name: str,
    save_or_display: List[Literal["save", "display"]],
):
    import matplotlib.pyplot as plt

    # Creates a copy of the graph
    H = G.copy()

    # Checks all the edges and removes some based on corr_direction
    for stock1, stock2, weight in G.edges(data=True):
        # if we only want to see the positive correlations we then delete the edges with weight smaller than 0
        if corr_direction == "positive":
            # it adds a minimum value for correlation.
            # If correlation weaker than the min, then it deletes the edge
            if weight["weight"] < 0 or weight["weight"] < min_correlation:
                H.remove_edge(stock1, stock2)
        # this part runs if the corr_direction is negative and removes edges with weights equal or largen than 0
        else:
            # it adds a minimum value for correlation.
            # If correlation weaker than the min, then it deletes the edge
            if weight["weight"] >= 0 or weight["weight"] > min_correlation:
                H.remove_edge(stock1, stock2)

    # crates a list for edges and for the weights
    edges, weights = zip(*nx.get_edge_attributes(H, "weight").items())

    # increases the value of weights, so that they are more visible in the graph
    weights = tuple([(1 + abs(x)) ** 2 for x in weights])

    # calculates the degree of each node
    # d = nx.degree(H)
    # creates list of nodes and a list their degrees that will be used later for their sizes
    # nodelist, node_sizes = zip(*d.items())

    # positions
    positions = nx.circular_layout(H)

    # Figure size
    plt.figure(figsize=(15, 15))

    # draws nodes
    nx.draw_networkx_nodes(
        H,
        positions,
        node_color="#DA70D6",
        # nodelist=nodelist,
        # the node size will be now based on its degree
        node_size=5000,
        # node_size=tuple([x**3 for x in node_sizes]),
        alpha=0.8,
    )

    # Styling for labels
    nx.draw_networkx_labels(H, positions, font_size=21, font_family="sans-serif")

    # edge colors based on weight direction
    if corr_direction == "positive":
        edge_colour = plt.cm.GnBu
    else:
        edge_colour = plt.cm.PuRd

    # draws the edges
    nx.draw_networkx_edges(
        H,
        positions,
        edgelist=edges,
        style="solid",
        # adds width=weights and edge_color = weights
        # so that edges are based on the weight parameter
        # edge_cmap is for the color scale based on the weight
        # edge_vmin and edge_vmax assign the min and max weights for the width
        width=weights,
        edge_color=weights,
        edge_cmap=edge_colour,
        edge_vmin=min(weights),
        edge_vmax=max(weights),
    )

    plt.axis("off")

    if "save" in save_or_display:
        plt.savefig(f"{save_location}/{file_name}_{corr_direction}.png", format="PNG")
    if "display" in save_or_display:
        if not get_global_state().run_type == RunType.test:
            plt.show()


def create_animation(
    path: str = "output/structured_data/plots", file_suffix: str = "positive"
):
    import os

    import matplotlib.pyplot as plt
    from matplotlib.animation import FuncAnimation

    files = os.listdir(path)

    nframes = len(files)
    plt.subplots_adjust(top=1, bottom=0, left=0, right=1)

    def animate(i: int):
        im = plt.imread(f"{path}/{str(i)}_{file_suffix}.png")
        plt.imshow(im)

    anim = FuncAnimation(
        plt.gcf(), animate, frames=nframes, interval=(2000.0 / nframes)
    )
    anim.save(f"{path}/output.gif", writer="imagemagick")



import pandas as pd
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    cohen_kappa_score,
    f1_score,
    log_loss,
    matthews_corrcoef,
    precision_score,
    recall_score,
)

from krisi.evaluate.group import Group
from krisi.evaluate.library.benchmarking import RandomClassifier, model_benchmarking
from krisi.evaluate.library.diagrams import (
    callibration_plot,
    display_density_plot,
    display_single_value,
    display_time_series,
)
from krisi.evaluate.library.metric_wrappers import (
    bennet_s,
    brier_multi,
    pred_y_imbalance_ratio,
    wrap_avg_precision,
    wrap_brier_score,
    wrap_roc_auc,
    y_label_imbalance_ratio,
)
from krisi.evaluate.metric import Metric
from krisi.evaluate.type import Calculation, MetricCategories, Purpose

accuracy_binary = Metric[float](
    name="Accuracy",
    key="accuracy",
    category=MetricCategories.class_err,
    info="In multilabel classification, this function computes subset accuracy: the set of labels predicted for a sample must exactly match the corresponding set of labels in y_true. https://scikit-learn.org/stable/modules/generated/sklearn.metrics.accuracy_score.html",
    func=accuracy_score,
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.objective,
)
""" ~ """
accuracy_binary_balanced = Metric[float](
    name="Balanced Accuracy",
    key="balanced_accuracy",
    category=MetricCategories.class_err,
    info="Compute the balanced accuracy. The balanced accuracy in binary and multiclass classification problems to deal with imbalanced datasets. It is defined as the average of recall obtained on each class.",
    func=balanced_accuracy_score,
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.objective,
)
""" ~ """

recall_binary, recall_macro = [
    Metric[float](
        name=f"Recall ({mode})",
        key=f"recall_{mode}",
        category=MetricCategories.class_err,
        info="The recall is the ratio tp / (tp + fn) where tp is the number of true positives and fn the number of false negatives. The recall is intuitively the ability of the classifier to find all the positive samples.\nThe best value is 1 and the worst value is 0. https://scikit-learn.org/stable/modules/generated/sklearn.metrics.recall_score.html",
        parameters={"average": mode},
        func=recall_score,
        plot_funcs=[(display_single_value, dict(width=750.0))],
        plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
        purpose=Purpose.objective,
    )
    for mode in ["binary", "macro"]
]
"""~"""
precision_binary, precision_macro = [
    Metric[float](
        name=f"Precision ({mode})",
        key=f"precision_{mode}",
        category=MetricCategories.class_err,
        info="The precision is the ratio tp / (tp + fp) where tp is the number of true positives and fp the number of false positives. The precision is intuitively the ability of the classifier not to label as positive a sample that is negative.\nThe best value is 1 and the worst value is 0. https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html",
        parameters={"average": mode},
        func=precision_score,
        plot_funcs=[(display_single_value, dict(width=750.0))],
        plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
        purpose=Purpose.objective,
    )
    for mode in ["binary", "macro"]
]

kappa = Metric[float](
    name="Cohen's Kappa",
    key="kappa",
    category=MetricCategories.class_err,
    info="Compute Cohen’s kappa: a statistic that measures inter-annotator agreement. This function computes Cohen’s kappa [1], a score that expresses the level of agreement between two annotators on a classification problem. ",
    func=cohen_kappa_score,
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.objective,
)

"""~"""
f_one_score_binary, f_one_score_macro, f_one_score_micro, f_one_score_weighted = [
    Metric[float](
        name=f"F1 Score ({mode})",
        key=f"f_one_score_{mode}",
        category=MetricCategories.class_err,
        info="The F1 score can be interpreted as a harmonic mean of the precision and recall, where an F1 score reaches its best value at 1 and worst score at 0. The relative contribution of precision and recall to the F1 score are equal. https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html",
        parameters={"average": mode},
        func=f1_score,
        plot_funcs=[(display_single_value, dict(width=750.0))],
        plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
        supports_multiclass=True,
        purpose=Purpose.objective,
    )
    for mode in ["binary", "macro", "micro", "weighted"]
]
"""~"""
matthew_corr = Metric[float](
    name="Matthew Correlation Coefficient",
    key="matthew_corr",
    category=MetricCategories.class_err,
    info="The Matthews correlation coefficient is used in machine learning as a measure of the quality of binary and multiclass classifications. It takes into account true and false positives and negatives and is generally regarded as a balanced measure which can be used even if the classes are of very different sizes. The MCC is in essence a correlation coefficient value between -1 and +1. A coefficient of +1 represents a perfect prediction, 0 an average random prediction and -1 an inverse prediction. The statistic is also known as the phi coefficient.",
    func=matthews_corrcoef,
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.objective,
)
"""~"""
brier_score = Metric[float](
    name="Brier Score",
    key="brier_score",
    category=MetricCategories.class_err,
    info="The smaller the Brier score loss, the better, hence the naming with “loss”. The Brier score measures the mean squared difference between the predicted probability and the actual outcome. The Brier score always takes on a value between zero and one, since this is the largest possible difference between a predicted probability (which must be between zero and one) and the actual outcome (which can take on values of only 0 and 1). It can be decomposed as the sum of refinement loss and calibration loss.",
    func=wrap_brier_score,
    parameters=dict(pos_label=1),
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    accepts_probabilities=True,
    supports_multiclass=False,
    purpose=Purpose.loss,
)
"""~"""
calibration = Metric[float](
    name="Calibration Plot",
    key="calibration",
    category=MetricCategories.class_err,
    info="Used to plot the calibration of a model with it probabilities.",
    func=lambda y, prob, **kwargs: pd.concat(
        [y, prob.iloc[:, 0].rename("probs")], axis="columns"
    ),
    plot_funcs=[(callibration_plot, dict(width=1500.0, bin_size=0.1))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    accepts_probabilities=True,
    supports_multiclass=True,
    calculation=Calculation.single,
    purpose=Purpose.diagram,
)
"""~"""
brier_score_multi = Metric[float](
    name="Brier Score Multilabel",
    key="brier_score_multi",
    category=MetricCategories.class_err,
    info="Multilabel calculation of the Brier score loss. The smaller the Brier score loss, the better, hence the naming with “loss”. The Brier score measures the mean squared difference between the predicted probability and the actual outcome. The Brier score always takes on a value between zero and one, since this is the largest possible difference between a predicted probability (which must be between zero and one) and the actual outcome (which can take on values of only 0 and 1). It can be decomposed as the sum of refinement loss and calibration loss.",
    func=lambda y, prob, **kwargs: brier_multi(y, prob, **kwargs),
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    accepts_probabilities=True,
    supports_multiclass=True,
    purpose=Purpose.loss,
)
"""~"""
s_score = Metric[float](
    name="S Score",
    key="s_score",
    category=MetricCategories.class_err,
    info="the 𝑆 score first proposed by Bennett, Alpert, & Goldstein (1954). It assumes a 50% probability of classifying any given item into its correct class 'by chance' alone and discounts this from the final score. It also uses all cells of the confusion matrix (unlike 𝐹1 which ignores 𝑑 or the number of 'true negatives')",
    func=bennet_s,
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    accepts_probabilities=False,
    supports_multiclass=False,
    calculation=Calculation.single,
    purpose=Purpose.objective,
)
"""~"""


cross_entropy = Metric[float](
    name="Cross Entropy",
    key="cross_entropy",
    category=MetricCategories.class_err,
    info="Log loss, aka logistic loss or cross-entropy loss. This is the loss function used in (multinomial) logistic regression and extensions of it such as neural networks, defined as the negative log-likelihood of a logistic model that returns y_pred probabilities for its training data y_true.",
    func=lambda y, prob, **kwargs: log_loss(
        y, prob, labels=list(prob.columns), **kwargs
    ),
    plot_funcs=[
        (display_single_value, dict(width=750.0)),
        (display_density_plot, dict(width=1500.0)),
    ],
    plot_funcs_rolling=[
        (display_density_plot, dict(width=1500.0)),
        (display_time_series, dict(width=1500.0)),
    ],
    accepts_probabilities=True,
    supports_multiclass=True,
    purpose=Purpose.loss,
)
"""~"""

roc_auc_binary_micro, roc_auc_binary_macro, roc_auc_binary_weighted = [
    Metric[float](
        name=f"ROC AUC ({mode}))",
        key=f"roc_auc_binary_{mode}",
        category=MetricCategories.class_err,
        info="Compute Area Under the Receiver Operating Characteristic Curve (ROC AUC) from prediction scores. Note: this implementation can be used with binary, multiclass and multilabel classification, but some restrictions apply (see Parameters). https://scikit-learn.org/stable/modules/generated/sklearn.metrics.roc_auc_score.html",
        func=wrap_roc_auc,
        parameters={"average": mode},
        plot_funcs=[(display_single_value, dict(width=750.0))],
        plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
        accepts_probabilities=True,
        supports_multiclass=True,
        purpose=Purpose.objective,
        calculation=Calculation.single,
    )
    for mode in ["micro", "macro", "weighted"]
]
"""~"""


roc_auc_multi_micro, roc_auc_multi_macro = [
    Metric[float](
        name=f"ROC AUC ({mode}))",
        key=f"roc_auc_multi_{mode}",
        category=MetricCategories.class_err,
        info="Compute Area Under the Receiver Operating Characteristic Curve (ROC AUC) from prediction scores. Note: this implementation can be used with binary, multiclass and multilabel classification, but some restrictions apply (see Parameters). https://scikit-learn.org/stable/modules/generated/sklearn.metrics.roc_auc_score.html",
        func=wrap_roc_auc,
        parameters={"average": mode, "multi_class": "ovr"},
        plot_funcs=[(display_single_value, dict(width=750.0))],
        plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
        accepts_probabilities=True,
        supports_multiclass=True,
        purpose=Purpose.objective,
        calculation=Calculation.single,
    )
    for mode in ["micro", "macro"]
]

avg_precision_micro, avg_precision_macro, avg_precision_weighted = [
    Metric[float](
        name=f"Average Precision ({mode}))",
        key=f"avg_precision_{mode}",
        category=MetricCategories.class_err,
        info="Compute average precision (AP) from prediction scores. AP summarizes a precision-recall curve as the weighted mean of precisions achieved at each threshold, with the increase in recall from the previous threshold used as the weight. https://scikit-learn.org/stable/modules/generated/sklearn.metrics.average_precision_score.html#sklearn.metrics.average_precision_score",
        func=wrap_avg_precision,
        parameters={"average": mode},
        plot_funcs=[(display_single_value, dict(width=750.0))],
        plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
        accepts_probabilities=True,
        supports_multiclass=True,
        purpose=Purpose.objective,
    )
    for mode in ["micro", "macro", "weighted"]
]
"""~"""
# """~"""
# ndcg = Metric[float](
#     name="NDCG Score",
#     key="ndcg",
#     category=MetricCategories.class_err,
#     info="Compute Normalized Discounted Cumulative Gain. Sum the true scores ranked in the order induced by the predicted scores, after applying a logarithmic discount. Then divide by the best possible score (Ideal DCG, obtained for a perfect ranking) to obtain a score between 0 and 1. This ranking metric returns a high value if true labels are ranked high by y_score.",
#     func=ndcg_score,
#     parameters={},
#     plot_funcs=[(display_single_value, dict(width=750.0))],
#     plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
#     accepts_probabilities=False,
#     supports_multiclass=True,
# )

imbalance_ratio_y = Metric[float](
    name="Imbalance Ratio: labels in y",
    key="imbalance_ratio_y",
    category=MetricCategories.class_err,
    info="Measurement of how skewed the target dataset is",
    func=y_label_imbalance_ratio,
    parameters={"pos_label": 1},
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    accepts_probabilities=False,
    supports_multiclass=False,
    purpose=Purpose.diagram,
    calculation=Calculation.single,
)
imbalance_ratio_pred_y = Metric[float](
    name="Imbalance Ratio: y vs pred",
    key="imbalance_ratio_pred_y",
    category=MetricCategories.class_err,
    info="Measurement of how skewed the pos_label for prediction vs target is",
    func=pred_y_imbalance_ratio,
    parameters={"pos_label": 1},
    plot_funcs=[(display_single_value, dict(width=750.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    accepts_probabilities=False,
    supports_multiclass=False,
    purpose=Purpose.diagram,
    calculation=Calculation.single,
)


binary_classification_balanced_metrics = [
    avg_precision_macro,
    accuracy_binary,
    recall_binary,
    precision_binary,
    f_one_score_binary,
    kappa,
    brier_score,
    calibration,
    roc_auc_binary_macro,
    imbalance_ratio_y,
    imbalance_ratio_pred_y,
]

binary_classification_imbalanced_metrics = [
    avg_precision_macro,
    accuracy_binary_balanced,
    recall_binary,
    precision_binary,
    f_one_score_binary,
    kappa,
    brier_score,
    roc_auc_binary_macro,
    imbalance_ratio_y,
    imbalance_ratio_pred_y,
]
binary_classification_metrics_balanced_benchmarking = [
    Group[pd.Series](
        name="benchmark",
        key="benchmark",
        metrics=binary_classification_balanced_metrics,
        postprocess_funcs=[model_benchmarking(RandomClassifier())],
        append_key=False,
    )
]
binary_classification_metrics_imbalanced_benchmarking = [
    Group[pd.Series](
        name="benchmark",
        key="benchmark",
        metrics=binary_classification_imbalanced_metrics,
        postprocess_funcs=[model_benchmarking(RandomClassifier())],
        append_key=False,
    )
]


"""~"""
minimal_binary_classification_metrics = [accuracy_binary, f_one_score_binary]
"""~"""
multiclass_classification_metrics = [
    cross_entropy,
    recall_macro,
    precision_macro,
    f_one_score_weighted,
    f_one_score_micro,
    brier_score_multi,
    roc_auc_multi_macro,
    # roc_auc_multi_weighted,
    matthew_corr,
    # ndcg,
    # s_score,
]
"""~"""
minimal_multiclass_classification_metrics = [
    f_one_score_weighted,
    # roc_auc_multi_weighted,
]



from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, List, Tuple, TypeVar, Union

import numpy as np
import pandas as pd

from krisi.utils.enums import ParsableEnum

MetricResult = TypeVar(
    "MetricResult", bound=Union[float, int, str, List, Tuple, pd.DataFrame, pd.Series]
)


Predictions = Union[np.ndarray, pd.Series, List[Union[int, float]]]
Targets = Union[np.ndarray, pd.Series, List[Union[int, float]]]
Probabilities = Union[np.ndarray, pd.Series, pd.DataFrame, List[Union[int, float]]]
Weights = Union[np.ndarray, pd.Series, List[Union[int, float]]]
PredictionsDS = pd.Series
ProbabilitiesDF = pd.DataFrame
TargetsDS = pd.Series
WeightsDS = pd.Series
MetricFunctionProb = Callable[[PredictionsDS, TargetsDS, ProbabilitiesDF], MetricResult]
MetricFunctionProbWeights = Callable[
    [PredictionsDS, TargetsDS, ProbabilitiesDF, WeightsDS], MetricResult
]
MetricFunctionNoProb = Callable[[PredictionsDS, TargetsDS], MetricResult]
MetricFunctionNoProbWeights = Callable[
    [PredictionsDS, TargetsDS, WeightsDS], MetricResult
]
MetricFunction = Union[
    MetricFunctionNoProb,
    MetricFunctionProb,
    MetricFunctionProbWeights,
    MetricFunctionNoProbWeights,
]


class MetricCategories(ParsableEnum):
    residual = "Residual Diagnostics"
    entropy = "Information Entropy"
    class_err = "Forecast Errors - Classification"
    reg_err = "Forecast Errors - Regression"
    stats = "General Statistics"
    unknown = "Unknown"


class SampleTypes(ParsableEnum):
    insample = "insample"
    validation = "validation"
    outofsample = "outofsample"


class Calculation(ParsableEnum):
    single = "single"
    rolling = "rolling"
    both = "both"


class SaveModes(ParsableEnum):
    svg = "svg"
    html = "html"
    text = "text"
    obj = "obj"
    minimal = "minimal"


class PathConst:
    default_eval_output_path: Path = Path("output/")
    default_analyse_output_path: Path = Path("analyse/")
    default_save_output_path: Path = Path("scorecards/")
    html_report_template_url: Path = Path("library/pdf_layouts/scorecard/report.html")
    css_report_template_url: Path = Path("library/pdf_layouts/scorecard/report.css")


class ComputationalComplexity(ParsableEnum):
    low = "low"
    medium = "medium"
    high = "high"


@dataclass
class ScoreCardMetadata:
    save_path: Path
    project_name: str = ""
    project_description: str = ""
    model_name: str = ""
    model_description: str = ""
    dataset_name: str = ""
    dataset_description: str = ""

    def __setattr__(self, key: str, item: Any) -> None:
        if key == "project_name":
            self.save_path = Path(
                os.path.join(PathConst.default_eval_output_path, item)
            )

        super().__setattr__(key, item)


class PrintMode(ParsableEnum):
    extended = "extended"
    minimal = "minimal"
    minimal_table = "minimal_table"


class NamingPrefixes:
    model = "Model_"
    dataset = "Dataset_"
    project = "Project_"


class DatasetType(ParsableEnum):
    regression = "regression"
    classification_binary_balanced = "classification_binary_balanced"
    classification_binary_imbalanced = "classification_binary_imbalanced"
    classification_multiclass = "classification_multiclass"

    def is_classification(self):
        return self in [
            DatasetType.classification_binary_balanced,
            DatasetType.classification_binary_imbalanced,
            DatasetType.classification_multiclass,
        ]


class Purpose(ParsableEnum):
    objective = "objective"
    loss = "loss"
    diagram = "diagram"
    group = "group"



import pandas as pd


def __check_gaps(df: pd.DataFrame, time_delta: pd.Timedelta) -> pd.Series:
    mask = df.index.to_series().diff() == time_delta
    return mask[mask == False]


def __check_frequency(df: pd.DataFrame) -> str:
    return pd.infer_freq(df.index)


def __check_nans(df: pd.DataFrame):
    return df[df.isna()]


def check_consistency(
    df: pd.DataFrame,
    title: str = "",
    time_delta: pd.Timedelta = pd.Timedelta("00:60:00"),
):
    """
    Prints a summary on:

    * the number of NaNs
    * the frequency of the dataset
    * if there are irregular gaps within the indexes.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame that we would like to check the inconsistency of.
    title : str
        Title of the report
    time_delta: pd.Timedelta
        The size of gaps within indexes, to check irregular indexes.

    """
    print(f"-- Consistency for {title} --")
    print(f"Gaps should be 1: {len(__check_gaps(df, time_delta))}")
    print(f"Frequency should be 'H': {__check_frequency(df)}")
    print(f"These are the NaN values: {__check_nans(df)}")
    print("----------------------------")



from importlib.util import find_spec

from krisi.utils.devutils.environment_checks import is_testing
from krisi.utils.state import GlobalState, RunType, set_global_state

if is_testing():
    set_global_state(GlobalState(run_type=RunType.test))
else:
    set_global_state(GlobalState(run_type=RunType.dev))



from . import library as library
from .compare import compare
from .metric import Metric
from .score import score
from .scorecard import ScoreCard
from .type import Calculation, MetricCategories, SampleTypes



from functools import reduce
from pathlib import Path
from typing import List, Optional, Tuple

import pandas as pd
from typing_extensions import Literal

from krisi.evaluate.type import PathConst
from krisi.report.graph import create_save_graphs
from krisi.utils.state import RunType, get_global_state

from .utils import corr_without_symmetry, unroll


def get_mean_corr_matrix(df_rolled_corr: List[pd.DataFrame]) -> pd.DataFrame:
    df_rolled_summed = reduce(lambda x, y: x.add(y, fill_value=0), df_rolled_corr)
    df_rolled_mean = df_rolled_summed / len(df_rolled_corr)

    return df_rolled_mean


def get_unrolled_correlation_over_time(
    df_rolled_corr: List[pd.DataFrame], start_indecies: List[pd.DatetimeIndex]
) -> pd.DataFrame:
    df_unrolled_in_time = pd.concat(
        [unroll(df_, start_indecies[i]) for i, df_ in enumerate(df_rolled_corr)],
        axis="columns",
    ).T
    df_unrolled_in_time.index = pd.to_datetime(df_unrolled_in_time.index)
    return df_unrolled_in_time


def get_corr_rolled(
    df: pd.DataFrame, window: int, step: Optional[int] = None
) -> Tuple[List[pd.DataFrame], List[pd.DataFrame]]:
    if step is None or step == 1 or step == 0:
        df_rolled = list(df.rolling(window=window))[1:]
    else:
        df_rolled = list(df.rolling(window=window, step=step))[1:]
    df_rolled_corr = [corr_without_symmetry(df_) for df_ in df_rolled]

    return df_rolled_corr, df_rolled


def __corr_mean(df_rolled_corr: List[pd.DataFrame], threshold: float):
    # Mean correlations in a matrix for plotting into a network graph
    mean_corr_over_time = get_mean_corr_matrix(df_rolled_corr)

    mean_corr_over_time_treshold = mean_corr_over_time.where(
        mean_corr_over_time >= threshold, other=0.0
    )

    return mean_corr_over_time, mean_corr_over_time_treshold


def __create_summary(
    corr_std: pd.Series,
    corr_mean: pd.Series,
    df_unrolled_in_time: pd.DataFrame,
    window: int,
    step: int,
    name: str,
    save_or_display: List[Literal["save", "display"]],
    save_location: str,
    top_k: int,
) -> None:
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(2, 1)
    title = f"Summary of {name} | window - {window} ~ step - {step}"
    fig.suptitle(title)

    # ranking = corr_mean/corr_std

    df_unrolled_in_time[corr_std.nsmallest(top_k).index.values].plot(
        title=f"Correlation over time (for top-{top_k} smallest standard deviation)",
        ax=ax[0],
        xticks=[],
    )
    df_unrolled_in_time[corr_mean.nlargest(top_k).index.values].plot(
        title=f"Correlation over time (for top-{top_k} largest mean)",
        ax=ax[1],
    )

    result = pd.concat([corr_std, corr_mean], axis="columns", keys=["std", "mean"])
    result.index.name = title
    print(result)

    if "save" in save_or_display:
        path_friendly = title.replace(" ", "_")
        plt.savefig(f"{save_location}/{path_friendly}.png", format="PNG")
    if "display" in save_or_display:
        if not get_global_state().run_type == RunType.test:
            plt.show()


def get_rolled_corr_metrics(
    df: pd.DataFrame,
    window: int,
    step: int = 1,
) -> Tuple[pd.DataFrame, pd.Series, pd.Series]:
    df_rolled_corr, df_rolled = get_corr_rolled(df, window, step)

    start_indecies = [df_.index[0] for df_ in df_rolled]

    coor_unrolled_in_time = get_unrolled_correlation_over_time(
        df_rolled_corr, start_indecies
    )

    # Correlations properties on unrolled correlation data
    corr_std = coor_unrolled_in_time.std()
    corr_mean = coor_unrolled_in_time.mean()

    return (
        coor_unrolled_in_time,
        corr_std,
        corr_mean,
    )


def matrix_corr_over_time(
    df: pd.DataFrame,
    window: int,
    step: int = 1,
    threshold: float = 0.25,
    save_location: Path = PathConst.default_analyse_output_path,
    save_graphs: bool = False,
):
    df_rolled_corr, df_rolled = get_corr_rolled(df, window, step)
    df_rolled_mean, df_rolled_mean_treshold = __corr_mean(
        df_rolled_corr, threshold=threshold
    )

    if save_graphs:
        create_save_graphs([df_rolled_mean_treshold], save_location)

    return df_rolled_mean, df_rolled_mean_treshold


def plot_corr_over_time(
    df: pd.DataFrame,
    window: int,
    step: int = 1,
    name: str = "",
    save_or_display: List[Literal["save", "display"]] = ["display"],
    save_location: str = "output/structured_data/plots",
    top_k: int = 2,
) -> None:
    (
        coor_unrolled_in_time,
        corr_std,
        corr_mean,
    ) = get_rolled_corr_metrics(df, window, step)

    __create_summary(
        corr_std,
        corr_mean,
        coor_unrolled_in_time,
        window,
        step,
        name=name,
        save_or_display=save_or_display,
        save_location=save_location,
        top_k=top_k,
    )



from __future__ import annotations

import datetime
import logging
import os
from copy import deepcopy
from dataclasses import asdict
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import pandas as pd
from rich import print
from typing_extensions import Literal

from krisi.evaluate.assertions import (
    check_no_duplicate_metrics,
    check_valid_pred_target,
    infer_dataset_type,
)
from krisi.evaluate.group import Group
from krisi.evaluate.library import get_default_metrics_for_dataset_type
from krisi.evaluate.metric import Metric
from krisi.evaluate.type import (
    DatasetType,
    MetricCategories,
    PathConst,
    Predictions,
    PredictionsDS,
    PrintMode,
    Probabilities,
    ProbabilitiesDF,
    SampleTypes,
    SaveModes,
    ScoreCardMetadata,
    Targets,
    TargetsDS,
    Weights,
    WeightsDS,
)
from krisi.evaluate.utils import (
    convert_to_series,
    ensure_df,
    get_save_path,
    handle_unnamed,
    rename_probs_columns,
)
from krisi.report.console import (
    get_large_metric_summary,
    get_minimal_summary,
    get_summary,
)
from krisi.report.report import create_report_from_scorecard
from krisi.report.type import DisplayModes, InteractiveFigure
from krisi.utils.environment import is_notebook
from krisi.utils.io import ensure_path, save_console, save_minimal_summary, save_object
from krisi.utils.iterable_helpers import (
    empty_if_none,
    flatten,
    map_newdict_on_olddict,
    remove_nans,
    replace_if_none,
    strip_builtin_functions,
    wrap_in_list,
)
from krisi.utils.state import RunType, get_global_state, set_global_state

logger = logging.getLogger("krisi")


class ScoreCard:
    """
    ScoreCard Object.

    Krisi's main object holding and evaluating metrics.
    Stores, evaluates and generates vizualisations of predefined
    and custom metrics for regression and classification.

    Parameters
    -------
    y: Targets
        True Targets to which the metrics are evaluated to.
    predictions: Predictions
        The single point predictions to which the metrics are evaluated to.
    model_name: Optional[str]
        The name of the model that the predictions were generated by. Used for identifying scorecards.
    model_description: str
        A description of the model that the predictions were generated by. Used for reporting.
    dataset_name: Optional[str]
        The name of the dataset from which the `y` (targets) orginate from. Used for reporting.
    dataset_description: str = ""
        A description of the dataset from which the `y` (targets) orginate from. Used for reporting.
    project_name: Optional[str]
        The name of the project. Used for reporting and saving to a directory (eg.: multiple scorecards)
    project_description: str = ""
        A description of the project. Used for reporting.
    dataset_type: Optional[Union[DatasetType, str]]
        Whether the task was a binar/multi-label classifiction of regression. If set to `None` it will infer from the target.
    sample_type: SampleTypes = SampleTypes.outofsample
        Whether we should evaluate it on insample or out of sample.

            - `SampleTypes.outofsample`
            - `SampleTypes.insample`
    default_metrics: Optional[List[Metric]]
        Default metrics that get evaluated. See `library`.
    custom_metrics: Optional[List[Metric]]
        Custom metrics that get evaluated. If specified it will evaluate these after `default_metric`
        See `library`.
    rolling_args : Dict[str, Any], optional
        Arguments to be passed onto `pd.DataFrame.rolling`.
        Default:

        - The window size of the rolling metric evaluation. If `None` evaluation over time will be on expanding window basis, by default `None`.
        - The step size of the rolling metric evaluation, by default `1`.

    Examples
    --------
    >>> from krisi import ScoreCard
    ... y_pred, y_true = [0, 2, 1, 3], [0, 1, 2, 3]
    ... sc = ScoreCard()
    ... sc.evaluate(y_pred, y_true, defaults=True) # Calculate predefined metrics
    ... sc["own_metric"] = (y_pred - y_true).mean() # Add a metric result directly
    ... sc.print()
    """

    y: TargetsDS
    predictions: PredictionsDS
    probabilities: Optional[ProbabilitiesDF]
    sample_weight: Optional[WeightsDS]
    sample_type: SampleTypes
    default_metrics_keys: List[str]
    custom_metrics_keys: List[str]
    dataset_type: DatasetType
    metadata: ScoreCardMetadata
    rolling_args: Dict[str, Any]

    def __init__(
        self,
        y: Targets,
        predictions: Predictions,
        probabilities: Optional[Probabilities] = None,
        sample_weight: Optional[Weights] = None,
        model_name: Optional[str] = None,
        model_description: str = "",
        dataset_name: Optional[str] = None,
        dataset_description: str = "",
        project_name: Optional[str] = None,
        project_description: str = "",
        dataset_type: Optional[Union[DatasetType, str]] = None,
        sample_type: Union[str, SampleTypes] = SampleTypes.outofsample,
        default_metrics: Optional[Union[List[Metric], Metric]] = None,
        custom_metrics: Optional[Union[List[Metric], Metric]] = None,
        rolling_args: Optional[Dict[str, Any]] = None,
        raise_exceptions: bool = False,
    ) -> None:
        if raise_exceptions:
            state = get_global_state()
            state.run_type = RunType.test
            set_global_state(state)

        sample_type = SampleTypes.from_str(sample_type)
        dataset_type = DatasetType.from_str(dataset_type) if dataset_type else None
        check_valid_pred_target(y, predictions, sample_weight)
        default_metrics = (
            wrap_in_list(default_metrics) if default_metrics is not None else None
        )
        custom_metrics = (
            wrap_in_list(custom_metrics) if custom_metrics is not None else None
        )

        check_no_duplicate_metrics(
            empty_if_none(default_metrics) + empty_if_none(custom_metrics)
        )
        self.__dict__["y"] = convert_to_series(y, "y")
        self.__dict__["predictions"] = convert_to_series(predictions, "predictions")
        self.__dict__["probabilities"] = (
            rename_probs_columns(ensure_df(probabilities, "probabilities"))
            if probabilities is not None
            else None
        )
        self.__dict__["sample_weight"] = (
            convert_to_series(
                sample_weight,  # pd.Series([1.0] * len(y))
                "sample_weight",
            )
            if sample_weight is not None
            else None
        )

        self.__dict__["sample_type"] = sample_type
        if rolling_args is None:
            window_size = len(y) // 20  # 5% of the data
            rolling_args = dict(
                window=window_size,
                step=window_size,
                min_periods=window_size,
                closed="left",
            )
            logger.info(f"rolling_args not specified, using default: {rolling_args}")

        self.__dict__["rolling_args"] = (
            rolling_args if rolling_args is not None else dict(window=len(y) // 100)
        )
        self.__dict__["dataset_type"] = (
            infer_dataset_type(y) if dataset_type is None else dataset_type
        )
        model_name_, dataset_name_, project_name_ = handle_unnamed(
            y, predictions, model_name, dataset_name, project_name
        )

        self.__dict__["metadata"] = ScoreCardMetadata(
            get_save_path(project_name_),
            project_name_,
            project_description,
            model_name_,
            model_description,
            dataset_name_,
            dataset_description,
        )

        if default_metrics is None:
            self.__dict__["dataset_type"] = infer_dataset_type(y)
            default_metrics = get_default_metrics_for_dataset_type(self.dataset_type)

        self.__dict__["default_metrics_keys"] = [
            metric.key for metric in default_metrics
        ]
        if custom_metrics is None:
            custom_metrics = []

        self.__dict__["custom_metrics_keys"] = [metric.key for metric in custom_metrics]

        for metric in default_metrics:
            self.__dict__[metric.key] = deepcopy(metric)

        for metric in custom_metrics:
            self.__dict__[metric.key] = deepcopy(metric)

    def __setitem__(self, key: str, item: Any) -> None:
        self.__setattr__(key, item)

    def __getitem__(self, key: Union[str, List[str]]) -> Union[ScoreCard, Metric]:
        if isinstance(key, List):
            scorecard_copy = deepcopy(self)
            all_keys = list(scorecard_copy.__dict__.keys())
            for k in all_keys:
                if k not in key:
                    if isinstance(scorecard_copy.__dict__[k], Metric):
                        del scorecard_copy.__dict__[k]

            return scorecard_copy
        elif isinstance(key, str):
            if "-" in key:
                # We'd like to get a comparison result
                splitted_key = key.split("-")
                assert len(splitted_key) == 2, f"Key {key} is not valid"
                metric_key, comparison_key = splitted_key

                metric = deepcopy(getattr(self, metric_key))

                assert isinstance(
                    metric.comparison_result, pd.Series
                ), f"Metric {metric_key} does not have comparison results"
                return Metric(
                    **{
                        **asdict(metric.reset()),
                        **dict(
                            name=f"{metric.name}-{comparison_key}",
                            key=f"{metric.key}-{comparison_key}",
                            result=metric.comparison_result[comparison_key],
                        ),
                    },
                )

            else:
                return deepcopy(getattr(self, key))

    def __delitem__(self, key: str) -> None:
        del self[key]

    def __setattr__(self, key: str, item: Any) -> None:
        """
        Defines Dictionary like behaviour and ensures that a Metric can be
        added as a:

            - `Metric` object,
            - Dictionary,
            - Direct result (float, int or a List of float or int). Gets wrapped in a `Metric` object

        Parameters
        ----------
        key : string
            The key to which the object will be assigned to.
        item : Dictionary, Metric, Float, Int or List of Float or Int, or pd.Series
            The result that gets stored with the key. Depending on the type of object it will result in
            different behaviours:

                - If `Metric` or `Dict` it will store the object on `ScoreCard[key]`
                - If (`Float`, `Int`, `List of Float`, `List of Int`, `pd.Series`) it will check if a `Metric` on `key`
                already exists on `ScoreCard`. If yes, it will assign item to the result field of the
                existing `Metric`. If not it will wrap the item in a new `Metric` first then assign it to
                `ScoreCard[key]`.

        Examples
        --------
        >>> from krisi import ScoreCard
        ... sc = ScoreCard()
        ... sc['metric_result'] = 0.53 # Direct result assignment as a Dictionary
        Metric(result=0.53, key='metric_result', category=None, parameters=None, info="", ...)

        >>> sc.another_metric_result = 1 # Direct object assignment
        Metric(result=1, key='another_metric_result', category=None, parameters=None, info="", ...)

        >>> from krisi.evaluate.metric import Metric
        ... from krisi.evaluate.type import MetricCategories
        ... sc.full_metric = Metric("My own metric", category=MetricCategories.class_err, info="A fictious metric with metadata", func: lambda y, y_hat: (y - y_hat)/2)
        Metric("My own metric", key="my_own_metric", category=MetricCategories.class_err, info="A fictious metric with metadata", ...)

        >>> sc.metric_as_dictionary = {name: "My other metric", info: "A Metric created with a dictionary", func: lambda y, y_hat: y - y_hat}
        Metric("My other metric", key="my_other_metric", info="A Metric created with a dictionary", ...)

        """
        metric = getattr(self, key, None)

        if metric is None:
            if isinstance(item, dict):
                self.__dict__[key] = Metric(key=key, **item)
            elif isinstance(item, Metric):
                item["key"] = key
                self.__dict__[key] = item
            else:
                self.__dict__[key] = Metric(
                    name=key, key=key, result=item, category=MetricCategories.unknown
                )
        else:
            if isinstance(item, dict):
                metric_dict = map_newdict_on_olddict(
                    strip_builtin_functions(vars(metric)), item, exclude=["name"]
                )
                self.__dict__[key] = Metric(**metric_dict)
            elif isinstance(item, Metric):
                metric_dict = map_newdict_on_olddict(
                    strip_builtin_functions(vars(metric)),
                    strip_builtin_functions(vars(item)),
                    exclude=["name"],
                )
                self.__dict__[key] = Metric(**metric_dict)
            else:
                metric["result"] = item
                self.__dict__[key] = metric

    def get_ds(self, name_as_index: bool = False) -> pd.Series:
        """
        Returns a `pd.Series` where each index is the name of a `Metric` and
        the value is the corresponding `result`

        Returns
        -------
        pd.Series
        """
        metrics = [
            metric
            for metric in self.get_all_metrics()
            if not isinstance(metric.result, Iterable) and not isinstance(metric, Group)
        ]

        comparison_results = [
            metric["comparison_result"].add_prefix(f"{metric.key}-")
            for metric in metrics
            if isinstance(metric["comparison_result"], pd.Series)
        ]
        if len(comparison_results) == 0:
            comparison_results = pd.Series([])
        else:
            comparison_results = pd.concat(comparison_results, axis=0)

        return pd.concat(
            [
                pd.Series(
                    [metric["result"] for metric in metrics],
                    index=[
                        metric.name if name_as_index else metric.key
                        for metric in metrics
                    ],
                ),
                comparison_results,
            ]
        )

    def get_default_metrics(self) -> List[Metric]:
        """
        Returns a List of Predefined Metrics according to task type:
        regression, classification, multi-label classification.

        Returns
        -------
        List of Metrics
        """
        return [
            self.__dict__[key]
            for key in self.default_metrics_keys
            if key in self.__dict__
        ]

    def get_custom_metrics(self) -> List[Metric]:
        """
        Returns a List of Custom ``Metric``s defined by the user on initalization
        of the ``ScoreCard``

        Returns
        -------
        List of Metrics
        """
        predifined_custom_metrics = [
            self.__dict__[key]
            for key in self.custom_metrics_keys
            if key in self.__dict__
        ]
        modified_custom_metrics = [
            value
            for key, value in self.__dict__.items()
            if key not in self.custom_metrics_keys + self.default_metrics_keys
            and hasattr(value, "key")
        ]

        return predifined_custom_metrics + modified_custom_metrics

    def get_all_metrics(
        self,
        defaults: bool = True,
        only_evaluated: bool = False,
        spread_comparisons: bool = False,
    ) -> List[Metric]:
        """
        Helper function that returns both `default_metrics` and `custom_metrics`.

        Parameters
        -------
        only_evaluated: bool
            Only return `Metric`s that were already evaluated
        defauls: bool
            Whether `default_metrics` should be return or not.

        Returns
        -------
        List of Metrics
        """
        if defaults:
            metrics = self.get_default_metrics() + self.get_custom_metrics()
        else:
            metrics = self.get_custom_metrics()

        if spread_comparisons:
            metrics += [
                Metric(
                    **{
                        **asdict(metric.reset()),
                        **dict(
                            name=f"{metric.name}-{index}",
                            key=f"{metric.key}-{index}",
                            result=result,
                        ),
                    },
                )
                for metric in metrics
                if isinstance(metric["comparison_result"], pd.Series)
                for index, result in metric["comparison_result"].items()
            ]

        if only_evaluated:
            return [metric for metric in metrics if metric.is_evaluated()]
        else:
            return metrics

    def __evaluate(
        self,
        func_key_evaluate: Literal["evaluate", "evaluate_over_time"],
        result_key: Literal["result", "result_rolling"],
        defaults: bool = True,
        rolling_args: Optional[Dict[str, Any]] = dict(),
    ):
        for metric in self.get_all_metrics(defaults=defaults):
            if metric.restrict_to_sample is not self.sample_type:
                getattr(metric, func_key_evaluate)(
                    self.y,
                    self.predictions,
                    self.probabilities,
                    self.sample_weight,
                    **rolling_args,
                )

    def evaluate_rolling_properties(self):
        for metric in self.get_all_metrics(defaults=True):
            metric.evaluate_rolling_properties()

    def evaluate(self, defaults: bool = True) -> None:
        """
        Evaluates `Metric`s present on the `ScoreCard`

        Parameters
        ----------
        defaults: boolean
            Wether the default `Metric`s should be evaluated or not.

        Returns
        -------
        None
        """
        self.__evaluate("evaluate", "result", defaults=defaults)

    def evaluate_over_time(self, defaults: bool = True) -> None:
        """
        Evaluates `Metric`s present on the `ScoreCard` over time, either with expanding
        or fixed sized window. Assigns list of results to `results_over_time`.

        Parameters
        ----------
        y: Targets = Union[np.ndarray, pd.Series, List[Union[int, float]]]
            The true labels to compare values to.
        predictions: Predictions = Union[np.ndarray, pd.Series, List[Union[int, float]]]
            The predicted values. Integers or whole floats if classification, else floats.
        defaults: bool
            Wether the default ``Metric``s should be evaluated or not. Default value = True
        window: int
            Size of window. If number is provided then evaluation happens on a fixed window size,
            otherwise it evaluates it on an expanding window basis.

        Returns
        -------
        None
        """
        self.__evaluate(
            "evaluate_over_time",
            "result_rolling",
            defaults=defaults,
            rolling_args={"rolling_args": self.rolling_args},
        )
        self.evaluate_rolling_properties()

    def print(
        self,
        mode: Union[str, PrintMode, List[PrintMode], List[str]] = PrintMode.extended,
        with_info: bool = False,
        with_parameters: bool = True,
        with_diagnostics: bool = False,
        input_analysis: bool = True,
        title: Optional[str] = None,
        frame_or_series: bool = True,
    ) -> None:
        """
        Prints the ScoreCard to the console.

        Parameters
        ----------
        mode: Union[str, PrintMode, List[PrintMode], List[str]] = PrintMode.extended
            - `PrintMode.extended` or `'extended'` prints the full ScoreCard, with targets, predictions, residuals, etc.
            - `PrintMode.minimal` or `'minimal'` prints the name and the matching result of each metric in the ScoreCard, without fancy formatting
            - `PrintMode.minimal_table` or `'minimal_table'` creates a table format of just the metric name and the accompanying result
        with_info: bool
            Wether descriptions of each metrics should be printed or not
        input_analysis: bool = True
            Wether it should print analysis about the raw `targets` and `predictions`
        title: Optional[str] = None
            Title of the table when mode = `'minimal_table'`
        """

        modes = [PrintMode.from_str(mode_) for mode_ in wrap_in_list(mode)]
        if title is None:
            title = self.metadata.project_name
        for mode in modes:
            if mode is PrintMode.extended:
                to_display = get_summary(
                    self,
                    repr=True,
                    categories=[el.value for el in MetricCategories],
                    with_info=with_info,
                    with_parameters=with_parameters,
                    with_diagnostics=with_diagnostics,
                    input_analysis=input_analysis,
                )

            elif mode is PrintMode.minimal:
                to_display = get_minimal_summary(self, dataframe=frame_or_series)
            elif mode is PrintMode.minimal_table:
                to_display = get_large_metric_summary(
                    self,
                    title,
                    with_info=with_info,
                    with_parameters=with_parameters,
                    with_diagnostics=with_diagnostics,
                )
            if isinstance(to_display, (pd.DataFrame, pd.Series)) and is_notebook():
                from IPython.display import display

                display(to_display)
            else:
                print(to_display)

    def cleanup_group(self) -> None:
        for metric in self.get_all_metrics(defaults=True):
            if isinstance(metric, Group):
                group = metric
                for metric_in_group in group.get_all_metrics():
                    if metric_in_group.key not in self.__dict__:
                        metric_in_group._from_group = True
                        self[metric_in_group.key] = metric_in_group
                    else:
                        self[metric_in_group.key] = {
                            f"result_{group.key}": metric_in_group.__dict__["result"],
                            f"result_rolling_{group.key}": metric_in_group.__dict__[
                                "result_rolling"
                            ],
                            "_from_group": True,
                        }

    def save(
        self,
        with_info: bool = False,
        with_parameters: bool = True,
        with_diagnostics: bool = False,
        save_modes: List[Union[SaveModes, str]] = [
            SaveModes.minimal,
            SaveModes.obj,
            SaveModes.text,
        ],
        override_base_path: Optional[Path] = None,
        timestamp_no_overwrite: bool = False,
    ) -> ScoreCard:
        if override_base_path is None:
            if timestamp_no_overwrite is False:
                dir_model_name = self.metadata.model_name
            else:
                dir_model_name = datetime.datetime.now().strftime("%H-%M-%S-%f")

        path = replace_if_none(
            override_base_path,
            os.path.join(
                self.metadata.save_path,
                os.path.join(
                    PathConst.default_save_output_path,
                    Path(dir_model_name),
                ),
            ),
        )
        ensure_path(path)

        if SaveModes.minimal in save_modes or SaveModes.minimal.value in save_modes:
            save_minimal_summary(self, path)
        if SaveModes.obj in save_modes or SaveModes.obj.value in save_modes:
            save_object(self, path)
        save_console(
            self,
            path,
            with_info,
            with_parameters=with_parameters,
            with_diagnostics=with_diagnostics,
            save_modes=save_modes,
        )

        return self

    def get_diagram_dictionary(self) -> Dict[str, InteractiveFigure]:
        diagrams = flatten(
            remove_nans([metric.get_diagrams() for metric in self.get_all_metrics()])
        )

        return {diagram.id: diagram for diagram in diagrams}

    def generate_report(
        self,
        display_modes: Union[str, List[str], DisplayModes, List[DisplayModes]] = [
            DisplayModes.interactive
        ],
        html_template_url: Path = PathConst.html_report_template_url,
        css_template_url: Path = PathConst.css_report_template_url,
        author: str = "",
        report_title: Optional[str] = None,
        override_base_path: Optional[Path] = None,
    ) -> None:
        save_path = replace_if_none(override_base_path, self.metadata.save_path)
        ensure_path(save_path)

        display_modes = [
            DisplayModes.from_str(mode) for mode in wrap_in_list(display_modes)
        ]
        report = create_report_from_scorecard(
            self,
            display_modes,
            html_template_url,
            css_template_url,
            author=author,
            report_title=replace_if_none(
                report_title, f"Report on {self.metadata.project_name}"
            ),
            save_path=save_path,
        )
        report.generate_launch()

    def __repr__(self):
        md = self.__dict__["metadata"]
        return (
            super().__repr__()[:-1]
            + f"\n{md.model_name:>40s} | {md.dataset_name} \n{md.project_name:>40s} | {self.dataset_type.value}\n\n{get_minimal_summary(self, dataframe=False)}>"
        )

    def __union(
        self,
        other: ScoreCard,
        function: Callable,
    ) -> ScoreCard:
        copied_scorecard = deepcopy(self)

        for key, value in copied_scorecard.__dict__.items():
            if isinstance(value, Metric) and key in other.__dict__.keys():
                if isinstance(value.result, (int, float)) and isinstance(
                    other[key].result, (int, float)
                ):
                    new_metric = value.reset()
                    new_metric.result = function(value.result, other[key].result)
                    if isinstance(value.comparison_result, pd.Series) and isinstance(
                        other[key].comparison_result, pd.Series
                    ):
                        new_metric.comparison_result = function(
                            value.comparison_result,
                            other[key].comparison_result,
                        )
                    copied_scorecard.__dict__[key] = new_metric
                else:
                    copied_scorecard.__dict__[key] = copied_scorecard.__dict__[
                        key
                    ].reset()

        return copied_scorecard

    def subtract(self, other: ScoreCard) -> ScoreCard:
        return self.__union(other, lambda x, y: x - y)

    def subtract_abs(self, other: ScoreCard) -> ScoreCard:
        return self.__union(other, lambda x, y: abs(x - y))

    def add(self, other: ScoreCard) -> ScoreCard:
        return self.__union(other, lambda x, y: x + y)

    def multiply(self, other: ScoreCard) -> ScoreCard:
        return self.__union(other, lambda x, y: x * y)

    def divide(self, other: ScoreCard) -> ScoreCard:
        return self.__union(other, lambda x, y: x / y)


def get_rolling_diagrams(obj: ScoreCard) -> List[List[InteractiveFigure]]:
    return [
        diagram
        for diagram in [
            metric.get_diagram_over_time() for metric in obj.get_all_metrics()
        ]
        if diagram is not None
    ]



import logging
from typing import Any, Optional, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.metrics import (
    average_precision_score,
    brier_score_loss,
    confusion_matrix,
    roc_auc_score,
)

from krisi.evaluate.type import (
    Predictions,
    PredictionsDS,
    ProbabilitiesDF,
    TargetsDS,
    WeightsDS,
)

logger = logging.getLogger("krisi")


def ljung_box(
    y: TargetsDS, predictions: Predictions, **kwargs
) -> Union[pd.DataFrame, Tuple[Any, Any], Tuple[Any, Any, Any, Any]]:
    from statsmodels.stats.diagnostic import acorr_ljungbox

    return acorr_ljungbox(predictions)


def create_one_hot(ds: Union[TargetsDS, ProbabilitiesDF]) -> pd.DataFrame:
    return pd.get_dummies(ds)


def brier_multi(targets: TargetsDS, probs: ProbabilitiesDF, **kwargs):
    # See https://stats.stackexchange.com/questions/403544/how-to-compute-the-brier-score-for-more-than-two-classes
    return np.mean(
        np.sum(
            (probs - create_one_hot(targets)) ** 2,
            axis=1,
        )
    )


def wrap_brier_score(y: TargetsDS, probs: ProbabilitiesDF, **kwargs) -> float:
    return brier_score_loss(
        y_true=y,
        y_prob=probs.iloc[:, 1],
        **kwargs,
    )


def wrap_roc_auc(
    y: TargetsDS,
    probs: ProbabilitiesDF,
    sample_weight: Optional[None] = None,
    **kwargs,
) -> float:
    probs = probs.rename(columns={col: i for i, col in enumerate(probs.columns)})
    prob_columns = list(probs.columns)
    if len(prob_columns) == 2:
        probs = probs.iloc[:, 1]
    if len(np.unique(y)) == 1:
        return "ROC AUC only works with more than one class."

    return roc_auc_score(
        y_true=y,
        y_score=probs,
        labels=prob_columns,
        **kwargs,
    )


def wrap_avg_precision(
    y: TargetsDS,
    probs: ProbabilitiesDF,
    sample_weight: Optional[None] = None,
    **kwargs,
) -> float:
    probs = probs.rename(columns={col: i for i, col in enumerate(probs.columns)})
    prob_columns = list(probs.columns)
    if len(prob_columns) == 2:
        probs = probs.iloc[:, 1]

    return average_precision_score(
        y_true=y,
        y_score=probs,
        **kwargs,
    )


def bennet_s(
    y: TargetsDS,
    preds: PredictionsDS,
    sample_weight: WeightsDS,
    **kwargs,
) -> float:
    """Bennett, Alpert, & Goldstein S score.
    Notes
    -------
    See https://stats.stackexchange.com/a/217404/315248

    Returns
    -------
    float
        float
    """
    tn, fp, fn, tp = confusion_matrix(y, preds, sample_weight=sample_weight).ravel()

    p_0 = (tn + tp) / (tn + fp + fn + tp)

    S_Score = (2 * p_0) - 1
    return S_Score


def y_label_imbalance_ratio(
    y: TargetsDS,
    preds: PredictionsDS,
    sample_weight: Optional[WeightsDS] = None,
    pos_label: int = 1,
    **kwargs,
) -> float:
    label_frequencies = y.value_counts(normalize=True).dropna()
    perfect_balance = 0.5
    if len(label_frequencies) == 2:
        return abs(label_frequencies[pos_label] - perfect_balance)

    elif len(label_frequencies) == 1:
        return 0.0
    else:
        raise ValueError(
            f"Imbalance only works on binary classification problems, got more than 2 labels: {label_frequencies}"
        )


def pred_y_imbalance_ratio(
    y: TargetsDS,
    preds: PredictionsDS,
    sample_weight: Optional[WeightsDS] = None,
    pos_label: int = 1,
    **kwargs,
) -> float:
    perfect_balance = 0.5
    prediction_frequencies = preds.value_counts(normalize=True).dropna()
    target_frequencies = y.value_counts(normalize=True).dropna()
    if len(prediction_frequencies) == 1 or len(target_frequencies) == 1:
        return 0.0
    elif len(prediction_frequencies) != 2 or len(target_frequencies) != 2:
        raise ValueError(
            f"pred_y_imbalance_ratio only works on binary classification problems, got more than 2 labels for either preds: {prediction_frequencies} or target: {target_frequencies}"
        )

    return abs(
        abs(target_frequencies[pos_label] - perfect_balance)
        - abs(prediction_frequencies[pos_label] - perfect_balance)
    )



from copy import deepcopy
from typing import Any, Dict, Generic, List, Optional, Tuple, Union

from krisi.utils.iterable_helpers import wrap_in_list

from .metric import Metric, PostProcessFunction
from .type import (
    Calculation,
    MetricFunction,
    MetricResult,
    PredictionsDS,
    ProbabilitiesDF,
    Purpose,
    TargetsDS,
    WeightsDS,
)


def deepcopy_and_evaluate(
    postprocess_func: Metric, metric: Metric, sample_weight: WeightsDS, rolling: bool
) -> Metric:
    postprocess_metric = deepcopy(postprocess_func)
    postprocess_metric._evaluation(
        metric.result_rolling if rolling else metric.result, sample_weight=sample_weight
    )
    return postprocess_metric


class Group(Metric, Generic[MetricResult]):
    metrics: List[Metric]

    def __init__(
        self,
        name: str,
        key: str,
        metrics: List[Metric],
        calculation: Union[str, Calculation] = Calculation.both,
        preprocess_func: Optional[MetricFunction] = None,
        postprocess_funcs: Optional[
            Union[List[PostProcessFunction], PostProcessFunction]
        ] = None,
        purpose: Purpose = Purpose.group,
        append_key: bool = True,
    ) -> None:
        self.calculation = Calculation.from_str(calculation)
        self.preprocess_func = preprocess_func
        self.postprocess_funcs = (
            wrap_in_list(postprocess_funcs) if postprocess_funcs else None
        )
        self.key = key
        self.name = name
        self.metrics = deepcopy(metrics)
        self.purpose = purpose
        if append_key:
            for metric in self.metrics:
                metric.key = f"{metric.key}_{self.key}"

    def _preprocess(
        self,
        y: TargetsDS,
        predictions: PredictionsDS,
        probabilities: ProbabilitiesDF,
        sample_weights: WeightsDS,
    ) -> Optional[Tuple]:
        results = (
            self.preprocess_func(
                y, predictions, probabilities, sample_weights=sample_weights
            )
            if self.preprocess_func is not None
            else None
        )

        if not isinstance(results, (list, tuple)) and results is not None:
            results = (results,)
        return results

    def _postprocess(
        self,
        all_metrics: List[Metric],
        y: TargetsDS,
        predictions: PredictionsDS,
        probabilities: ProbabilitiesDF,
        sample_weight: WeightsDS,
        rolling: bool,
    ) -> List[Metric]:
        if self.postprocess_funcs is not None:
            all_metrics_ = []

            for postprocess_func in self.postprocess_funcs:
                all_metrics_ += postprocess_func(
                    all_metrics,
                    y,
                    predictions,
                    probabilities,
                    sample_weight=sample_weight,
                    rolling=rolling,
                )

            all_metrics = all_metrics_
        self.metrics = all_metrics

        return self.metrics

    def get_all_metrics(self) -> List[Metric]:
        return self.metrics

    def __handle_args_to_pass_in(
        self,
        result: Any,
        y: TargetsDS,
        predictions: PredictionsDS,
        probabilities: ProbabilitiesDF,
        accept_probabilities: bool,
    ):
        if result is None:
            if accept_probabilities:
                if probabilities is None:
                    return None
                else:
                    return [y, probabilities]
            else:
                return [y, predictions]
        else:
            return result

    def evaluate_rolling_properties(self):
        for metric in self.metrics:
            metric.evaluate_rolling_properties()

    def evaluate(
        self,
        y: TargetsDS,
        predictions: PredictionsDS,
        probabilities: ProbabilitiesDF,
        sample_weight: WeightsDS,
    ) -> List[Metric]:
        if self.calculation == Calculation.rolling:
            return []
        results = self._preprocess(y, predictions, probabilities, sample_weight)

        for metric in self.metrics:
            args_to_pass = self.__handle_args_to_pass_in(
                results,
                y,
                predictions,
                probabilities,
                metric.accepts_probabilities,
            )
            if args_to_pass is not None:
                metric._evaluation(
                    *args_to_pass,
                    sample_weight=sample_weight,
                )

        return self._postprocess(
            self.metrics,
            y,
            predictions,
            probabilities,
            sample_weight=sample_weight,
            rolling=False,
        )

    def evaluate_over_time(
        self,
        y: TargetsDS,
        predictions: PredictionsDS,
        probabilities: ProbabilitiesDF,
        sample_weight: WeightsDS,
        rolling_args: Dict[str, Any],
    ) -> List[Metric]:
        if self.calculation == Calculation.single:
            return []
        results = self._preprocess(y, predictions, probabilities, sample_weight)

        for metric in self.metrics:
            args_to_pass = self.__handle_args_to_pass_in(
                results,
                y,
                predictions,
                probabilities,
                metric.accepts_probabilities,
            )
            if args_to_pass is not None:
                metric._rolling_evaluation(
                    *args_to_pass,
                    sample_weight=sample_weight,
                    rolling_args=rolling_args,
                )

        return self._postprocess(
            self.metrics,
            y,
            predictions,
            probabilities,
            sample_weight=sample_weight,
            rolling=True,
        )

    def __str__(self) -> str:
        return " - ".join([metric.key for metric in self.metrics])



from typing import Any, Dict, List, Optional, Union

from krisi.evaluate.metric import Metric
from krisi.evaluate.scorecard import ScoreCard
from krisi.evaluate.type import (
    Calculation,
    DatasetType,
    Predictions,
    Probabilities,
    SampleTypes,
    Targets,
    Weights,
)


def score(
    y: Targets,
    predictions: Predictions,
    probabilities: Optional[Probabilities] = None,
    sample_weight: Optional[Weights] = None,
    model_name: Optional[str] = None,
    dataset_name: Optional[str] = None,
    project_name: Optional[str] = None,
    default_metrics: Optional[Union[List[Metric], Metric]] = None,
    custom_metrics: Optional[Union[List[Metric], Metric]] = None,
    dataset_type: Optional[Union[DatasetType, str]] = None,
    sample_type: Union[str, SampleTypes] = SampleTypes.outofsample,
    calculation: Union[Calculation, str] = Calculation.single,
    rolling_args: Optional[Dict[str, Any]] = None,
    raise_exceptions: bool = False,
    **kwargs,
) -> ScoreCard:
    """
    Creates a ScoreCard based on the passed in arguments, evaluates and then returns the ScoreCard.

    Parameters
    ----------
    y: Targets
        True Targets to which the metrics are evaluated to.
    predictions: Predictions
        The single point predictions to which the metrics are evaluated to.
    model_name: Optional[str]
        The name of the model that the predictions were generated by. Used for identifying scorecards.
    dataset_name: Optional[str]
        The name of the dataset from which the `y` (targets) orginate from. Used for reporting.
    project_name: Optional[str]
        The name of the project. Used for reporting and saving to a directory (eg.: multiple scorecards)
    default_metrics: Optional[List[Metric]]
        Default metrics that get evaluated. See `library`.
    custom_metrics: Optional[List[Metric]]
        Custom metrics that get evaluated. If specified it will evaluate these after `default_metric`
        See `library`.
    dataset_type: Optional[Union[DatasetType, str]]
        Whether the task was a binar/multi-label classifiction of regression. If set to `None` it will infer from the target.
    sample_type: SampleTypes = SampleTypes.outofsample
        Whether we should evaluate it on insample or out of sample.

            - `SampleTypes.outofsample`
            - `SampleTypes.insample`
    calculation : Union[Calculation, str], optional
        Whether it should evaluate `Metrics` on a rolling basis or on the whole prediction or both, by default Calculation.single

            - `Calculation.single`
            - `Calculation.rolling`
            - `Calculation.both`
    rolling_args : Dict[str, Any], optional
        Arguments to be passed onto `pd.DataFrame.rolling`.
        Default:

        - The window size of the rolling metric evaluation. If `None` evaluation over time will be on expanding window basis, by default `len(dataset)//100`.
        - The step size of the rolling metric evaluation, by default `len(dataset)//100`.

    Returns
    -------
    ScoreCard
        The ScoreCard Evaluated

    Raises
    ------
    ValueError
        If Calculation type is incorrectly specified.
    """

    calculation = Calculation.from_str(calculation)

    sc = ScoreCard(
        y=y,
        predictions=predictions,
        probabilities=probabilities,
        sample_weight=sample_weight,
        model_name=model_name,
        dataset_name=dataset_name,
        project_name=project_name,
        sample_type=sample_type,
        dataset_type=dataset_type,
        default_metrics=default_metrics,
        custom_metrics=custom_metrics,
        rolling_args=rolling_args,
        raise_exceptions=raise_exceptions,
        **kwargs,
    )

    if calculation == Calculation.single:
        sc.evaluate()
    elif calculation == Calculation.rolling:
        sc.evaluate_over_time()
    elif calculation == Calculation.both:
        sc.evaluate()
        sc.evaluate_over_time()

    else:
        raise ValueError(f"Calculation type {calculation} not recognized.")

    sc.cleanup_group()
    return sc



from typing import TYPE_CHECKING, List, Tuple

import numpy as np
import pandas as pd
from sklearn.calibration import calibration_curve

if TYPE_CHECKING:
    import plotly.graph_objects as go

from krisi.evaluate.type import MetricResult


def display_time_series(data: List[MetricResult], **kwargs) -> "go.Figure":
    import plotly.express as px

    title = kwargs.get("title", "")
    df = pd.DataFrame(data, columns=[title])
    df["iteration"] = list(range(len(data)))
    fig = px.line(
        df,
        x="iteration",
        y=title,
    )
    return fig


def display_single_value(data: MetricResult, **kwargs) -> "go.Figure":
    import plotly.graph_objects as go

    fig = go.Figure()

    fig.add_trace(
        go.Indicator(
            value=data,
            delta={},
            gauge={"axis": {"visible": False}},
            domain={"row": 0, "column": 0},
        )
    )

    return fig


def display_acf_plot(
    data: MetricResult, plot_pacf: bool = False, **kwargs
) -> "go.Figure":
    import plotly.graph_objects as go
    from statsmodels.tsa.stattools import acf, pacf

    title = "Partial Autocorrelation (PACF)" if plot_pacf else "Autocorrelation (ACF)"

    if not isinstance(data, pd.Series):
        data = pd.Series(data)

    corr_array = (
        pacf(data.dropna(), alpha=0.05) if plot_pacf else acf(data.dropna(), alpha=0.05)
    )
    lower_y = corr_array[1][:, 0] - corr_array[0]
    upper_y = corr_array[1][:, 1] - corr_array[0]

    fig = go.Figure()
    [
        fig.add_scatter(
            x=(x, x), y=(0, corr_array[0][x]), mode="lines", line_color="#3f3f3f"
        )
        for x in range(len(corr_array[0]))
    ]
    fig.add_scatter(
        x=np.arange(len(corr_array[0])),
        y=corr_array[0],
        mode="markers",
        marker_color="#1f77b4",
        marker_size=12,
    )
    fig.add_scatter(
        x=np.arange(len(corr_array[0])),
        y=upper_y,
        mode="lines",
        line_color="rgba(255,255,255,0)",
    )
    fig.add_scatter(
        x=np.arange(len(corr_array[0])),
        y=lower_y,
        mode="lines",
        fillcolor="rgba(32, 146, 230,0.3)",
        fill="tonexty",
        line_color="rgba(255,255,255,0)",
    )
    fig.update_traces(showlegend=False)
    fig.update_xaxes(range=[-1, len(corr_array[0]) + 2])
    fig.update_yaxes(zerolinecolor="#000000")

    fig.update_layout(title=title)
    return fig


def display_density_plot(data: MetricResult, **kwargs) -> "go.Figure":
    import plotly.express as px

    if not isinstance(data, pd.Series):
        data = pd.Series(data)
    fig = px.histogram(data, marginal="box")  # or violin, rug

    return fig


def calculate_calibration_bins(
    data: MetricResult, pos_label: int, n_bins: int, **kwargs
) -> Tuple[np.ndarray, np.ndarray]:
    y_true = data["y"]
    y_prob = data["probs"]

    prob_true, prob_pred = calibration_curve(
        y_true, y_prob, pos_label=pos_label, n_bins=n_bins
    )
    return prob_true, prob_pred


def callibration_plot(
    data: MetricResult, n_bins: int = 8, pos_label: int = 1, **kwargs
) -> "go.Figure":
    import plotly.graph_objects as go

    fraction_positives, bins = calculate_calibration_bins(data, pos_label, n_bins)
    scatter = go.Scatter(
        x=bins, y=fraction_positives, mode="lines+markers", name="Data Points"
    )

    perfect_calibration = go.Scatter(
        x=[0, 1],
        y=[0, 1],
        mode="lines",
        name="Perfect Calibration",
        line=dict(color="black", dash="dash"),
    )

    layout = go.Layout(
        title="Calibration Curve",
        xaxis=dict(title="Predicted Probability", tickvals=bins),
        yaxis=dict(title="Fraction of Positives"),
        showlegend=True,
    )

    fig = go.Figure(data=[scatter, perfect_calibration], layout=layout)
    return fig



from __future__ import annotations

import logging
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Generic, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from krisi.evaluate.type import (
    Calculation,
    ComputationalComplexity,
    MetricCategories,
    MetricFunction,
    MetricResult,
    PredictionsDS,
    ProbabilitiesDF,
    Purpose,
    SampleTypes,
    TargetsDS,
    WeightsDS,
)
from krisi.report.console import get_metric_string
from krisi.report.type import InteractiveFigure, PlotDefinition, plotly_interactive
from krisi.utils.iterable_helpers import (
    check_iterable_with_number,
    isiterable,
    string_to_id,
    wrap_in_list,
)
from krisi.utils.state import RunType, get_global_state


@dataclass
class Metric(Generic[MetricResult]):
    """
    Class representing a metric.

    Parameters
    ----------
    name: str
        The name of the metric.
    key: str
        The key used to reference the metric.
    category: MetricCategories
        The category of the metric.
    result: Optional[Union[Exception, MetricResult, List[MetricResult]]]
        The result of the evaluated `Metric`, by default None
    result_rolling: Optional[Union[Exception, MetricResult, List[MetricResult]]]
        The result of the evaluated `Metric` over time, by default None
    parameters: dict
        The paramaters that are passed into the evaluation function (param: `func`), by default field(default_factory=dict)
    func: Callable
        The function used to compute the metric.
    plot_funcs: List[Callable]
        List of functions used to plot the metric.
    plot_funcs_rolling: Callable
        Function used to plot the rolling metric value.
    info: str
        Additional information about the metric.
    restrict_to_sample: Optional[SampleTypes]
        Weather the `Metric` should only be evaluated on insample or out of sample data, by default None
    comp_complexity: Optional[ComputationalComplexity]
        How resource intensive the calculation is, by default None
    calculation: Calculation
        Whether the metric should be evaluated only when calculating rolling, single or both, by default both
    accepts_probabilities: bool
        Whether the metric accepts probabilities as input, by default False
    supports_multiclass: bool
        Whether the metric supports multiclass classification, by default False
    """

    name: str
    key: str = ""
    category: Optional[MetricCategories] = None
    result: Optional[Union[Exception, MetricResult, List[MetricResult]]] = None
    result_rolling: Optional[Union[Exception, MetricResult, List[MetricResult]]] = None
    rolling_properties: Optional[pd.Series] = None
    parameters: dict = field(default_factory=dict)
    func: Optional[MetricFunction] = None
    plot_funcs: Optional[Union[List[PlotDefinition], PlotDefinition]] = None
    plot_funcs_rolling: Optional[Union[List[PlotDefinition], PlotDefinition]] = None
    info: str = ""
    restrict_to_sample: Optional[SampleTypes] = None
    comp_complexity: Optional[ComputationalComplexity] = None
    calculation: Union[str, Calculation] = Calculation.both
    accepts_probabilities: bool = False
    supports_multiclass: bool = False
    diagnostics: Optional[Dict[str, Any]] = field(default_factory=dict)
    purpose: Optional[Union[str, Purpose]] = None
    comparison_result: Optional[pd.Series] = None
    _from_group: bool = False

    def __post_init__(self):
        if self.key == "":
            self.key = string_to_id(self.name)
        if not isinstance(self.plot_funcs, list) and self.plot_funcs is not None:
            self.plot_funcs = wrap_in_list(self.plot_funcs)
        if (
            not isinstance(self.plot_funcs_rolling, list)
            and self.plot_funcs_rolling is not None
        ):
            self.plot_funcs_rolling = wrap_in_list(self.plot_funcs_rolling)
        self.calculation = Calculation.from_str(self.calculation)
        self.purpose = (
            Purpose.from_str(self.purpose) if self.purpose is not None else None
        )

    def __setitem__(self, key: str, item: Any) -> None:
        setattr(self, key, item)

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key, "Unknown Field")

    def __str__(self, repr: bool = False) -> str:
        return get_metric_string(self, repr=repr)

    def __repr__(self) -> str:
        return super().__repr__()[:-1] + f" - {self.__str__(True)}>"

    def _evaluation(self, *args, **kwargs) -> Metric:
        if self.calculation == Calculation.rolling:
            return self
        if self._from_group:
            return self
        if self.func is None:
            raise ValueError("`func` has to be set on Metric to calculate result.")
        try:
            result = self.func(*args, **kwargs, **self.parameters)
        except Exception as e:
            result = e
            if get_global_state().run_type == RunType.test:
                raise e
        self.__safe_set(result, key="result")
        return self

    def evaluate(
        self,
        y: TargetsDS,
        predictions: PredictionsDS,
        probabilities: Optional[ProbabilitiesDF] = None,
        sample_weight: Optional[WeightsDS] = None,
    ) -> None:
        assert (
            self.func is not None
        ), "`func` has to be set on Metric to calculate result."
        if self.accepts_probabilities:
            if probabilities is not None:
                self._evaluation(y, probabilities, sample_weight=sample_weight)
            else:
                self.__safe_set(
                    ValueError(
                        "Metric requires probabilities, but None were provided."
                    ),
                    key="result",
                )
        else:
            self._evaluation(y, predictions, sample_weight=sample_weight)
        if sample_weight is not None:
            self.__dict__["diagnostics"] = dict(used_sample_weight=True)

    @staticmethod
    def __handle_window(
        window: pd.DataFrame,
    ) -> Tuple[Union[dict, Tuple], WeightsDS]:
        sample_weight = (
            window.pop("sample_weight") if "sample_weight" in window else None
        )
        if "y" not in window:
            return tuple([window[col] for col in window]), sample_weight

        y = window.pop("y")
        pred = window.pop("predictions") if "predictions" in window else None
        prob = window if len(window.columns) > 0 else None

        return dict(y=y, pred=pred, prob=prob), sample_weight

    @staticmethod
    def __calc_window(
        window: pd.DataFrame,
        func: MetricFunction,
        parameters: dict,
        accepts_probabilities: bool,
    ):
        result, sample_weight = Metric.__handle_window(window)

        if isinstance(result, dict):
            y = result["y"]
            pred = result["pred"]
            prob = result["prob"]

            if accepts_probabilities:
                if prob is not None:
                    return func(y, prob, sample_weight=sample_weight, **parameters)
                else:
                    return ValueError(
                        "Metric requires probabilities, but None were provided."
                    )
            else:
                return func(y, pred, sample_weight=sample_weight, **parameters)
        else:
            return func(
                *(list(result)),
                sample_weight=sample_weight,
                **parameters,
            )

    def _rolling_evaluation(self, *args, rolling_args: dict, **kwargs) -> Metric:
        if self._from_group:
            return self
        if self.calculation == Calculation.single:
            return self
        if self.func is None:
            raise ValueError("`func` has to be set on Metric to calculate result.")
        else:
            _df = pd.concat(
                kwargs.values() if (args is None or len(args) == 0) else args,
                axis="columns",
            )

            if "sample_weight" in _df:
                self.__dict__["diagnostics"] = dict(used_sample_weight=True)
            try:
                df_rolled = (
                    _df.expanding()
                    if "window" in rolling_args and rolling_args["window"] is None
                    else _df.rolling(**rolling_args)
                )

                result_rolling = [
                    Metric.__calc_window(
                        single_window,
                        self.func,
                        self.parameters,
                        self.accepts_probabilities,
                    )
                    for single_window in df_rolled
                    if len(single_window) > 0
                ]

            except Exception as e:
                result_rolling = e
                if get_global_state().run_type == RunType.test:
                    raise e
            self.__safe_set(result_rolling, key="result_rolling")

        return self

    def evaluate_over_time(
        self,
        y: TargetsDS,
        predictions: PredictionsDS,
        probabilities: Optional[ProbabilitiesDF] = None,
        sample_weight: Optional[WeightsDS] = None,
        rolling_args: dict = dict(),
    ) -> None:
        if self.accepts_probabilities and probabilities is not None:
            self._rolling_evaluation(
                y=y,
                predictions=predictions,
                probabilities=probabilities,
                sample_weight=sample_weight,
                rolling_args=rolling_args,
            )
        else:
            self._rolling_evaluation(
                y=y,
                predictions=predictions,
                sample_weight=sample_weight,
                rolling_args=rolling_args,
            )

    def evaluate_rolling_properties(self):
        if check_iterable_with_number(self.result_rolling):
            self.__safe_set(
                pd.Series(
                    data=dict(
                        mean=np.mean(self.result_rolling),
                        std=np.std(self.result_rolling),
                        min=np.min(self.result_rolling),
                        max=np.max(self.result_rolling),
                    )
                ),
                key="rolling_properties",
            )

    def is_evaluated(self, rolling: bool = False):
        if rolling:
            return self.result_rolling is not None
        else:
            return self.result is not None

    def get_diagram_over_time(self) -> Optional[List[InteractiveFigure]]:
        return create_diagram_rolling(self)

    def get_diagrams(
        self,
    ) -> Optional[List[InteractiveFigure]]:
        return create_diagram(self)

    def print(
        self,
        mode: Union[str, List[str]] = "dummy",
        with_info: bool = False,
        input_analysis: bool = True,
        title: Optional[str] = None,
    ) -> None:
        pass

    def __safe_set(
        self,
        result: Union[Exception, MetricResult, List[MetricResult], pd.Series],
        key: str,
    ):
        if self.__dict__[key] is not None:
            raise ValueError("This metric already contains a result.")
        else:
            self.__dict__[key] = result

    def reset(self, inplace: bool = False) -> Metric:
        if inplace:
            self.result = None
            self.result_rolling = None
            self.rolling_properties = None
            self.diagnostics = None
            self._from_group = False
            self.comparison_result = None
            return self
        else:
            return deepcopy(self).reset(inplace=True)


def create_diagram_rolling(obj: Metric) -> Optional[List[InteractiveFigure]]:
    if obj.plot_funcs_rolling is None:
        logging.info("No plot_funcs_rolling (Plotting Function Rolling) specified")
        return None
    elif isinstance(obj.result_rolling, Exception) or obj.result_rolling is None:
        return None
    elif isiterable(obj.result_rolling) and len(obj.result_rolling) > 0:
        if isinstance(obj.result_rolling[0], (pd.DataFrame, pd.Series)):
            return None
        return [
            InteractiveFigure(
                f"{obj.key}_{plot_funcs_rolling.__name__}",
                get_figure=plotly_interactive(
                    plot_funcs_rolling, obj.result_rolling, title=obj.name
                ),
                title=f"{obj.name} - {plot_funcs_rolling.__name__}",
                category=obj.category,
                plot_args=merge_default_dict(plot_args),
            )
            for plot_funcs_rolling, plot_args in obj.plot_funcs_rolling
        ]
    else:
        return None


def merge_default_dict(
    new_dict: Optional[Dict[str, Any]] = None,
    default_dict: Dict[str, Any] = dict(width=900.0, height=600.0),
) -> Dict[str, Any]:
    if new_dict is None:
        return default_dict
    else:
        return {**default_dict, **new_dict}


def create_diagram(obj: Metric) -> Optional[List[InteractiveFigure]]:
    if obj.plot_funcs is None:
        logging.info("No plot_func (Plotting Function) specified")
        return None
    elif isinstance(obj.result, Exception) or obj.result is None:
        return None
    else:
        return [
            InteractiveFigure(
                f"{obj.key}_{plot_func.__name__}",
                get_figure=plotly_interactive(plot_func, obj.result, title=obj.name),
                title=f"{obj.name} - {plot_func.__name__}",
                category=obj.category,
                plot_args=merge_default_dict(plot_args),
            )
            for plot_func, plot_args in obj.plot_funcs
        ]


PostProcessFunction = Callable[
    [
        List[Metric],
        TargetsDS,
        Optional[PredictionsDS],
        Optional[ProbabilitiesDF],
        Optional[WeightsDS],
        bool,
    ],
    Metric,
]



import os
from copy import deepcopy
from typing import TYPE_CHECKING, List, Tuple, Union

import numpy as np
import pandas as pd
from rich import box
from rich.pretty import Pretty
from rich.table import Table

from krisi.evaluate.type import Predictions, Targets
from krisi.report.library.console.diagrams import (
    callibration_plot,
    distribution_plot,
    histogram_plot,
    line_plot,
)
from krisi.utils.console_plot import plotextMixin
from krisi.utils.iterable_helpers import calculate_nans, filter_none, isiterable

if TYPE_CHECKING:
    from krisi.evaluate.metric import Metric


def bold(text: str, rich: bool = True) -> str:
    return f"[bold]{text}[/bold]" if rich else f"\033[1m{text}\033[0m"


def get_term_size() -> int:
    term_size = os.get_terminal_size()
    return term_size.columns


def __convert_result(
    result, title: str, return_string: bool = False
) -> Union[str, Pretty, plotextMixin]:
    result = deepcopy(result)
    if result is None:
        return ""
    if isinstance(result, float):
        result = round(result, 3)

        if return_string:
            return str(result)
    if isinstance(result, str) and return_string:
        return result
    elif isinstance(result, Exception):
        result = str(result)

    if isinstance(result, dict):
        if len(result) > 0:
            return "\n".join([f"{key}:\n {value}" for key, value in result.items()])

        else:
            return ""
    if isinstance(result, Tuple):

        def string_from_ds(ds: pd.Series, title: str) -> str:
            return "\n".join(
                [
                    f"({key}\n {__convert_result(value, title, True)})"
                    for key, value in ds.items()
                ]
            )

        result = filter_none(list(result))
        if len(result) == 1:
            return __convert_result(result[0], title, return_string=True)

        return "\n".join(
            [
                f"{__convert_result(res, title, return_string=True)}"
                if not isinstance(res, pd.Series)
                else string_from_ds(res, title)
                for res in result
                if res is not None
            ]
        )

    if isiterable(result):
        if isinstance(result, pd.DataFrame):
            if "probs" in result:
                return plotextMixin(result, callibration_plot, title=title)
            else:
                return Pretty(result)
        if isinstance(result, pd.Series):
            if len(result) > 10:
                return plotextMixin(result.tolist(), line_plot, title=title)
            else:
                return Pretty(result)
        elif isiterable(result[0]):
            return Pretty("Result is a complex Iterable")
        else:
            # Create a Console Plot
            return plotextMixin(result, line_plot, title=title)
    else:
        return Pretty(result, max_depth=2, max_length=3)


def __display_result(metric: "Metric") -> List[Union[str, Pretty, plotextMixin]]:
    return [
        __convert_result(res, title=metric.name)
        for res in [
            (metric.result, metric.comparison_result),
            metric.result_rolling,
            metric.rolling_properties,
        ]
    ]


def __create_metric(
    metric: "Metric", with_info: bool, with_parameters: bool, with_diagnostics: bool
) -> List[str]:
    results_as_strings = __display_result(metric)
    if isinstance(results_as_strings[0], plotextMixin):
        results_as_strings = ["", *results_as_strings[:-1]]
    metric_summarized = (
        [
            f"{metric.name} ({metric.key})",
            *results_as_strings,
        ]
        + ([__convert_result(metric.parameters, title="")] if with_parameters else [])
        + ([Pretty(metric.diagnostics)] if with_diagnostics else [])
        + ([Pretty(metric.info)] if with_info else [])
    )

    return metric_summarized


def create_metric_table(
    title: str,
    metrics: List["Metric"],
    with_info: bool,
    with_parameters: bool,
    with_diagnostics: bool,
    show_header: bool = False,
) -> Table:
    table = Table(
        title=title,
        # show_edge=False,
        show_footer=False,
        show_lines=True,
        show_header=show_header,
        expand=True,
        box=box.ROUNDED,
    )

    table.add_column(
        "Metric Name", justify="right", style="cyan", width=2, no_wrap=False
    )
    table.add_column("Result", style="magenta", width=1)
    table.add_column("Rolling", width=8)
    table.add_column("Rolling Props", width=1)
    if with_parameters:
        table.add_column("Parameters", style="green", width=1)
    if with_info:
        table.add_column("Info", width=3)
    if with_diagnostics:
        table.add_column("Diagnostics", width=3)

    for metric in metrics:
        if metric.result is None and metric.result_rolling is None:
            continue
        metric_summarized = __create_metric(
            metric, with_info, with_parameters, with_diagnostics
        )
        table.add_row(*metric_summarized)

    return table


def metrics_empty_in_category(metrics: List["Metric"]) -> bool:
    return (
        metrics is None
        or len(metrics) < 1
        or (
            all([metric.result is None for metric in metrics])
            and all([metric.result_rolling is None for metric in metrics])
        )
    )


def create_y_pred_table(classification: bool, y: Targets, preds: Predictions) -> Table:
    if isinstance(y, np.ndarray):
        y = pd.Series(y)
    if isinstance(preds, np.ndarray):
        preds = pd.Series(preds)

    table = Table(
        title="Targets and Predictions Analysis",
        # show_edge=False,
        show_footer=False,
        show_lines=True,
        show_header=True,
        expand=True,
        box=box.ROUNDED,
    )

    vizualisation_name = "Category Imbalance" if classification else "Histogram"
    vizualisation_func = distribution_plot if classification else histogram_plot
    table.add_column(
        "Series Type", justify="right", style="cyan", width=2, no_wrap=False
    )
    table.add_column(
        vizualisation_name, justify="left", style="cyan", width=10, no_wrap=False
    )
    table.add_column(
        "Types",
        justify="right",
        style="cyan",
        width=1,
        no_wrap=False,
    )
    table.add_column(
        "Indicies",
        justify="right",
        style="cyan",
        width=1,
        no_wrap=False,
    )

    table.add_row(
        "Targets",
        plotextMixin(y, vizualisation_func, title=vizualisation_name),
        f"NaNs: {str(calculate_nans(y))}\ndtype: {str(y.dtype)}",
        f"Start: {str(y.index[0])}\nEnd: {str(y.index[-1])}",
    )
    table.add_row(
        "Predictions",
        plotextMixin(preds, vizualisation_func, title=vizualisation_name),
        f"NaNs: {str(calculate_nans(preds))}\ndtype: {str(preds.dtype)}",
        f"Start: {str(preds.index[0])}\nEnd: {str(preds.index[-1])}",
    )
    return table



from typing import List, Optional, Union

import pandas as pd

from krisi.evaluate.scorecard import ScoreCard
from krisi.evaluate.utils import handle_empty_metrics_to_display
from krisi.utils.printing import bold


def compare(
    scorecards: List[ScoreCard],
    metric_keys: Optional[List[str]] = None,
    sort_by: Optional[str] = None,
    dataframe: bool = True,
) -> Union[pd.DataFrame, str]:
    """Creates a table where each column is a metric and each row is
    a scorecard and its corresponding results.

    Parameters
    ----------
    scorecards : List[ScoreCard]
        ScoreCards to compare.
    metric_keys : Optional[List[str]], optional
        List of metrics to display. If not set it will return all
        evaluated metrics on the first scorecard.
        Sorts the results by the first element of this list if `sort_by` is not specified., by default None
    sort_by : Optional[str], optional
        `Metric` to sort results by. Selected `Metric` will be displayed in the first row.
        If not specified metrics will be sorted by the first element of `metric_keys`.
        If `metric_keys` is not specified it will default to the first metric found on the first `ScoreCard`, by default None
    dataframe : bool, optional
        Whether it should return a `pd.DataFrame` or a `str`, by default True

    Returns
    -------
    Union[pd.DataFrame, str]
        A comparison table, either in `pd.DataFrame` or `string` format.
    """

    metric_keys, sort_by = handle_empty_metrics_to_display(
        scorecards[0], sort_by, metric_keys
    )

    scorecards.sort(reverse=True, key=lambda x: x[sort_by].result)
    if dataframe:
        return pd.concat(
            [
                pd.Series(
                    [scorecard[metric_key].result for scorecard in scorecards],
                    name=metric_key,
                    index=[scorecard.metadata.model_name for scorecard in scorecards],
                )
                for metric_key in metric_keys
            ],
            axis="columns",
        )
    else:
        metric_keys.remove(sort_by)
        string_to_return = ""
        metric_title = "".join([f"{metric:<10s}" for metric in metric_keys])
        string_to_return += f"{'model_name':>30s}    {bold(f'{sort_by:<10s}', rich=False)} {metric_title}"

        for scorecard in scorecards:
            metrics = "".join(
                [f"{scorecard[metric].result:<10.5}" for metric in metric_keys]
            )

            string_to_return += f"\n{scorecard.metadata.model_name:>30s}    {bold(f'{scorecard[sort_by].result:<10.5}', rich=False)} {metrics}"

        return string_to_return



from __future__ import annotations

from enum import Enum
from typing import Union

from typing_extensions import Self


class ParsableEnum(Enum):
    @classmethod
    def from_str(cls, value: Union[str, ParsableEnum]) -> Self:
        if isinstance(value, cls):
            return value
        for item in cls:
            if item.name == value:
                return item
        else:
            raise ValueError(f"Unknown {cls.__name__}: {value}")



from typing import TYPE_CHECKING, Dict, List

from krisi.utils.io import ensure_path

if TYPE_CHECKING:
    import plotly.graph_objects as go

from pathlib import Path

from krisi.report.type import InteractiveFigure


def __figure_to_html(figure: "go.Figure") -> str:
    import base64

    image = str(base64.b64encode(figure.to_image(format="svg")))[2:-1]
    return f'<img style="max-width:100%; max-height:100%;" src="data:image/png;base64,{image}"/>'


def convert_figure_to_html(figure: "go.Figure") -> str:
    return "<br>" + __figure_to_html(figure)


def convert_figures_to_html(figures: List["go.Figure"]) -> str:
    return "<br>".join([__figure_to_html(figure) for figure in figures])


def convert_figures_to_side_by_side_html(figures: List["go.Figure"]) -> str:
    return (
        '<div class="flex flex-row">'
        + "".join([f"<div>{__figure_to_html(figure)}</div>" for figure in figures])
        + "</div>"
    )


def __create_html_report(
    template_file: Path, html_elements_to_inject: Dict[str, str]
) -> str:
    import pkgutil

    template_html = pkgutil.get_data(__name__, str(template_file))
    template_html = template_html.decode()

    for key, value in html_elements_to_inject.items():
        template_html = template_html.replace(f"{{ {key} }}", value)
    return template_html


def __convert_html_to_pdf(
    source_html: Path, output_path: Path, report_name: str, css_template_url: Path
) -> None:
    pass

    # import pkgutil
    import logging

    from weasyprint import CSS, HTML

    logging.getLogger("weasyprint").setLevel(100)

    ensure_path(output_path)

    # css_string = pkgutil.get_data(__name__, str(css_template_url))
    # css_string = css_string.decode()

    # css = CSS(string=css_string)
    pdf = HTML(string=source_html).write_pdf(
        stylesheets=[
            CSS("https://cdn.jsdelivr.net/npm/tailwindcss/dist/tailwind.min.css")
        ]
    )

    open(f"{output_path}/{report_name}", "wb").write(pdf)


def create_pdf_report(
    path: Path,
    title: str,
    html_template_url: Path,
    css_template_url: Path,
    html_elements_to_inject: Dict[str, str] = dict(),
):
    html_elements_to_inject["title"] = title
    report_html = __create_html_report(html_template_url, html_elements_to_inject)
    __convert_html_to_pdf(report_html, path, f"{title}.pdf", css_template_url)


def append_sizes(
    diagram_dict: Dict[str, InteractiveFigure]
) -> Dict[str, InteractiveFigure]:
    size_dict = dict(
        residuals_display_acf_plot=(750.0, 750.0),
        residuals_display_density_plot=(350.0, 350.0),
        residuals_display_time_series=(1500.0, 750.0),
    )

    for key, value in size_dict.items():
        if key in diagram_dict.keys():
            # diagram_dict[key].width = value[0]
            diagram_dict[key].height = value[1]

    return diagram_dict



from typing import List

from krisi.evaluate.metric import Metric
from krisi.evaluate.type import DatasetType

from .default_metrics_classification import *
from .default_metrics_classification import (
    binary_classification_balanced_metrics,
    binary_classification_imbalanced_metrics,
    binary_classification_metrics_balanced_benchmarking,
    binary_classification_metrics_imbalanced_benchmarking,
    minimal_binary_classification_metrics,
    minimal_multiclass_classification_metrics,
    multiclass_classification_metrics,
)
from .default_metrics_regression import *
from .default_metrics_regression import (
    all_regression_metrics,
    low_computation_regression_metrics,
    minimal_regression_metrics,
)


def get_default_metrics_for_dataset_type(type: DatasetType) -> List[Metric]:
    if type == DatasetType.classification_binary_balanced:
        return binary_classification_metrics_balanced_benchmarking
    elif type == DatasetType.classification_binary_imbalanced:
        return binary_classification_metrics_imbalanced_benchmarking
    elif type == DatasetType.classification_multiclass:
        return multiclass_classification_metrics
    elif type == DatasetType.regression:
        return all_regression_metrics
    else:
        raise ValueError(f"Unknown dataset type {type}")



import math
from copy import deepcopy
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from krisi.evaluate.metric import Metric
from krisi.evaluate.type import (
    PredictionsDS,
    ProbabilitiesDF,
    Purpose,
    TargetsDS,
    WeightsDS,
)
from krisi.utils.data import generate_synthetic_predictions_binary, shuffle_df_in_chunks


class Model:
    def predict(
        self, X: pd.DataFrame, y: pd.Series, sample_weight: Optional[WeightsDS] = None
    ) -> Tuple[pd.Series, pd.DataFrame]:
        raise NotImplementedError


class RandomClassifier(Model):
    def __init__(self) -> None:
        self.name = "NS"

    def predict(
        self, X: pd.DataFrame, y: pd.Series, sample_weight: Optional[WeightsDS] = None
    ) -> Tuple[pd.Series, pd.DataFrame]:
        preds_probs = generate_synthetic_predictions_binary(y, sample_weight)
        predictions = preds_probs.iloc[:, 0]
        probabilities = preds_probs.iloc[:, 1:3]
        return predictions, probabilities


class RandomClassifierSmoothed(Model):
    def __init__(self, smoothing_window: Optional[int] = None) -> None:
        self.name = "NS-Smooth"
        self.smoothing_window = smoothing_window

    def predict(
        self, X: pd.DataFrame, y: pd.Series, sample_weight: Optional[WeightsDS] = None
    ) -> Tuple[pd.Series, pd.DataFrame]:
        def get_avarage_change_length(ds: pd.Series) -> float:
            # This only works for binary classification
            indices = ds.diff().fillna(0).ne(0).cumsum()
            differences = indices.groupby(indices).size()
            return differences.mean()

        smoothing_window = (
            math.floor(get_avarage_change_length(y))
            if self.smoothing_window is None
            else self.smoothing_window
        )

        preds_probs = generate_synthetic_predictions_binary(
            y, sample_weight, smoothing_window=smoothing_window
        )
        predictions = preds_probs.iloc[:, 0]
        probabilities = preds_probs.iloc[:, 1:3]
        return predictions, probabilities


class RandomClassifierChunked(Model):
    def __init__(self, chunk_size: Union[float, int]) -> None:
        self.name = "NS"
        self.chunk_size = chunk_size

    def predict(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weight: Optional[WeightsDS] = None,
    ) -> Tuple[pd.Series, pd.DataFrame]:
        X = shuffle_df_in_chunks(X, self.chunk_size)
        predictions = X.iloc[:, 0]
        probabilities = X.iloc[:, 1:3]
        return predictions, probabilities


class PerfectModel(Model):
    def __init__(self) -> None:
        self.name = "PM"

    def predict(
        self, X: pd.DataFrame, y: pd.Series, sample_weight: WeightsDS
    ) -> Tuple[pd.Series, pd.DataFrame]:
        predictions = y
        probabilities = pd.get_dummies(y, dtype=float)
        return predictions, probabilities


class WorstModel(Model):
    def __init__(self) -> None:
        self.name = "WM"

    def predict(
        self, X: pd.DataFrame, y: pd.Series, sample_weight: WeightsDS
    ) -> Tuple[pd.Series, pd.DataFrame]:
        predictions = np.logical_xor(y, 1).astype(int)
        probabilities = pd.get_dummies(predictions, dtype=float)
        return predictions, probabilities


def model_benchmarking(model: Model) -> Callable:
    def postporcess_func(
        all_metrics: List[Metric],
        y: TargetsDS,
        predictions: Optional[PredictionsDS],
        probabilities: Optional[ProbabilitiesDF],
        sample_weight: Optional[WeightsDS],
        rolling: bool,
    ) -> List[Metric]:
        if rolling:
            return all_metrics
        benchmark_predictions, benchmark_probabilities = model.predict(
            pd.concat([predictions, probabilities], axis="columns", copy=False),
            y,
            sample_weight,
        )

        for metric in all_metrics:
            if metric.purpose == Purpose.group or metric.purpose == Purpose.diagram:
                continue
            benchmark_metric = deepcopy(metric)
            benchmark_metric.result = None
            benchmark_metric.result_rolling = None

            if benchmark_metric.accepts_probabilities:
                benchmark_metric._evaluation(
                    y, benchmark_probabilities, sample_weight=sample_weight
                )
            else:
                benchmark_metric._evaluation(
                    y, benchmark_predictions, sample_weight=sample_weight
                )

            benchmark_result = (
                benchmark_metric.result
                if isinstance(benchmark_metric.result, (float, int))
                else None
            )
            model_result = (
                metric.result if isinstance(metric.result, (float, int)) else None
            )
            if benchmark_result is not None and model_result is not None:
                if metric.purpose == Purpose.objective:
                    comparison_result = model_result - benchmark_result
                elif metric.purpose == Purpose.loss:
                    comparison_result = benchmark_result - model_result
            else:
                comparison_result = "Either benchmark or model result is None."
            metric.comparison_result = pd.concat(
                [
                    metric.comparison_result,
                    pd.Series([comparison_result], index=[f"Δ {model.name}"]),
                ],
                axis=0,
            )
        return all_metrics

    return postporcess_func


if __name__ == "__main__":
    RandomClassifierSmoothed().predict(
        pd.Series([0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 0, 1]),
    )
